#!/usr/bin/env python
# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

#
# Typical usage:
# dohtml -r docs/*
#  - put all files and directories in docs into /usr/share/doc/${PF}/html
# dohtml foo.html
#  - put foo.html into /usr/share/doc/${PF}/html
#
#
# Detailed usage:
# dohtml <list-of-files>
#  - will install the files in the list of files (space-separated list) into
#    /usr/share/doc/${PF}/html, provided the file ends in .css, .gif, .htm,
#    .html, .jpeg, .jpg, .js or .png.
# dohtml -r <list-of-files-and-directories>
#  - will do as 'dohtml', but recurse into all directories, as long as the
#    directory name is not CVS
# dohtml -A jpe,java [-r] <list-of-files[-and-directories]>
#  - will do as 'dohtml' but add .jpe,.java (default filter list is
#    added to your list)
# dohtml -a png,gif,html,htm [-r] <list-of-files[-and-directories]>
#  - will do as 'dohtml' but filter on .png,.gif,.html,.htm (default filter
#    list is ignored)
# dohtml -x CVS,SCCS,RCS -r <list-of-files-and-directories>
#  - will do as 'dohtml -r', but ignore directories named CVS, SCCS, RCS
#

import os as _os
import sys

from portage import _unicode_encode, _unicode_decode, os, shutil
from portage.util import normalize_path, writemsg

# Change back to original cwd _after_ all imports (bug #469338).
os.chdir(os.environ["__PORTAGE_HELPER_CWD"])


def dodir(path):
    try:
        os.makedirs(path, 0o755)
    except OSError:
        if not os.path.isdir(path):
            raise
        os.chmod(path, 0o755)


def dofile(src, dst):
    shutil.copy(src, dst)
    os.chmod(dst, 0o644)


def eqawarn(lines):
    cmd = "source '%s/isolated-functions.sh' ; " % os.environ["PORTAGE_BIN_PATH"]
    for line in lines:
        cmd += 'eqawarn "%s" ; ' % line
    os.spawnlp(os.P_WAIT, "bash", "bash", "-c", cmd)


skipped_directories = []
skipped_files = []
warn_on_skipped_files = (
    os.environ.get("PORTAGE_DOHTML_WARN_ON_SKIPPED_FILES") is not None
)
unwarned_skipped_extensions = os.environ.get(
    "PORTAGE_DOHTML_UNWARNED_SKIPPED_EXTENSIONS", ""
).split()
unwarned_skipped_files = os.environ.get(
    "PORTAGE_DOHTML_UNWARNED_SKIPPED_FILES", ""
).split()


def install(basename, dirname, options, prefix=""):
    fullpath = basename
    if prefix:
        fullpath = os.path.join(prefix, fullpath)
    if dirname:
        fullpath = os.path.join(dirname, fullpath)

    if options.DOCDESTTREE:
        desttree = options.DOCDESTTREE
    else:
        desttree = "html"

    destdir = os.path.join(
        options.ED,
        "usr",
        "share",
        "doc",
        options.PF.lstrip(os.sep),
        desttree.lstrip(os.sep),
        options.doc_prefix.lstrip(os.sep),
        prefix,
    ).rstrip(os.sep)

    if not os.path.exists(fullpath):
        sys.stderr.write("!!! dohtml: %s does not exist\n" % fullpath)
        return False
    elif os.path.isfile(fullpath):
        ext = os.path.splitext(basename)[1][1:]
        if ext in options.allowed_exts or basename in options.allowed_files:
            dodir(destdir)
            dofile(fullpath, os.path.join(destdir, basename))
        elif (
            warn_on_skipped_files
            and ext not in unwarned_skipped_extensions
            and basename not in unwarned_skipped_files
        ):
            skipped_files.append(fullpath)
    elif (
        options.recurse
        and os.path.isdir(fullpath)
        and basename not in options.disallowed_dirs
    ):
        for i in _os.listdir(_unicode_encode(fullpath)):
            try:
                i = _unicode_decode(i, errors="strict")
            except UnicodeDecodeError:
                writemsg(
                    "dohtml: argument is not encoded as UTF-8: %s\n"
                    % _unicode_decode(i),
                    noiselevel=-1,
                )
                sys.exit(1)
            pfx = basename
            if prefix:
                pfx = os.path.join(prefix, pfx)
            install(i, dirname, options, pfx)
    elif not options.recurse and os.path.isdir(fullpath):
        global skipped_directories
        skipped_directories.append(fullpath)
        return False
    else:
        return False
    return True


class OptionsClass:
    def __init__(self):
        self.PF = ""
        self.ED = ""
        self.DOCDESTTREE = ""

        if "PF" in os.environ:
            self.PF = os.environ["PF"]
            if self.PF:
                self.PF = normalize_path(self.PF)
        if "force-prefix" not in os.environ.get(
            "FEATURES", ""
        ).split() and os.environ.get("EAPI", "0") in ("0", "1", "2"):
            self.ED = os.environ.get("D", "")
        else:
            self.ED = os.environ.get("ED", "")
        if self.ED:
            self.ED = normalize_path(self.ED)
        if "_E_DOCDESTTREE_" in os.environ:
            self.DOCDESTTREE = os.environ["_E_DOCDESTTREE_"]
            if self.DOCDESTTREE:
                self.DOCDESTTREE = normalize_path(self.DOCDESTTREE)

        self.allowed_exts = ["css", "gif", "htm", "html", "jpeg", "jpg", "js", "png"]
        self.allowed_files = []
        self.disallowed_dirs = ["CVS"]
        self.recurse = False
        self.verbose = False
        self.doc_prefix = ""


def print_help():
    opts = OptionsClass()

    print("dohtml [-a .foo,.bar] [-A .foo,.bar] [-f foo,bar] [-x foo,bar]")
    print("       [-r] [-V] <file> [file ...]")
    print()
    print(" -a   Set the list of allowed to those that are specified.")
    print("      Default:", ",".join(opts.allowed_exts))
    print(" -A   Extend the list of allowed file types.")
    print(" -f   Set list of allowed extensionless file names.")
    print(" -x   Set directories to be excluded from recursion.")
    print("      Default:", ",".join(opts.disallowed_dirs))
    print(" -p   Set a document prefix for installed files (empty by default).")
    print(" -r   Install files and directories recursively.")
    print(" -V   Be verbose.")
    print()


def parse_args():
    argv = sys.argv[:]

    # We can't trust that the filesystem encoding (locale dependent)
    # correctly matches the arguments, so use surrogateescape to
    # pass through the original argv bytes for Python 3.
    fs_encoding = sys.getfilesystemencoding()
    argv = [x.encode(fs_encoding, "surrogateescape") for x in argv]

    for x, arg in enumerate(argv):
        try:
            argv[x] = _unicode_decode(arg, errors="strict")
        except UnicodeDecodeError:
            writemsg(
                "dohtml: argument is not encoded as UTF-8: %s\n" % _unicode_decode(arg),
                noiselevel=-1,
            )
            sys.exit(1)

    options = OptionsClass()
    args = []

    x = 1
    while x < len(argv):
        arg = argv[x]
        if arg in ["-h", "-r", "-V"]:
            if arg == "-h":
                print_help()
                sys.exit(0)
            elif arg == "-r":
                options.recurse = True
            elif arg == "-V":
                options.verbose = True
        elif argv[x] in ["-A", "-a", "-f", "-x", "-p"]:
            x += 1
            if x == len(argv):
                print_help()
                sys.exit(0)
            elif arg == "-p":
                options.doc_prefix = argv[x]
                if options.doc_prefix:
                    options.doc_prefix = normalize_path(options.doc_prefix)
            else:
                values = argv[x].split(",")
                if arg == "-A":
                    options.allowed_exts.extend(values)
                elif arg == "-a":
                    options.allowed_exts = values
                elif arg == "-f":
                    options.allowed_files = values
                elif arg == "-x":
                    options.disallowed_dirs = values
        else:
            args.append(argv[x])
        x += 1

    return (options, args)


def main():

    (options, args) = parse_args()

    if options.verbose:
        print("Allowed extensions:", options.allowed_exts)
        print("Document prefix : '" + options.doc_prefix + "'")
        print("Allowed files :", options.allowed_files)

    success = False
    endswith_slash = (os.sep, os.sep + ".")

    for x in args:
        trailing_slash = x.endswith(endswith_slash)
        x = normalize_path(x)
        if trailing_slash:
            # Modify behavior of basename and dirname
            # as noted in bug #425214, causing foo/ to
            # behave similarly to the way that foo/*
            # behaves.
            x += os.sep
        basename = os.path.basename(x)
        dirname = os.path.dirname(x)
        success |= install(basename, dirname, options)

    for x in skipped_directories:
        eqawarn(["QA Notice: dohtml on directory '%s' without recursion option" % x])
    for x in skipped_files:
        eqawarn(["dohtml: skipped file '%s'" % x])

    if success:
        retcode = 0
    else:
        retcode = 1

    sys.exit(retcode)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2012-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

doc = """Dump and restore extended attributes.

We use formats like that used by getfattr --dump.  This is meant for shell
helpers to save/restore.  If you're looking for a python/portage API, see
portage.util.movefile._copyxattr instead.

https://en.wikipedia.org/wiki/Extended_file_attributes
"""
__doc__ = doc


import argparse
import array
import os
import re
import sys

from portage.util._xattr import xattr


_UNQUOTE_RE = re.compile(rb"\\[0-7]{3}")
_FS_ENCODING = sys.getfilesystemencoding()


def octal_quote_byte(b):
    return ("\\%03o" % ord(b)).encode("ascii")


def unicode_encode(s):
    if isinstance(s, str):
        s = s.encode(_FS_ENCODING, "surrogateescape")
    return s


def quote(s, quote_chars):
    """Convert all |quote_chars| in |s| to escape sequences

    This is normally used to escape any embedded quotation marks.
    """
    quote_re = re.compile(b"[" + quote_chars + b"]")
    result = []
    pos = 0
    s_len = len(s)

    while pos < s_len:
        m = quote_re.search(s, pos=pos)
        if m is None:
            result.append(s[pos:])
            pos = s_len
        else:
            start = m.start()
            result.append(s[pos:start])
            result.append(octal_quote_byte(s[start : start + 1]))
            pos = start + 1

    return b"".join(result)


def unquote(s):
    """Process all escape sequences in |s|"""
    result = []
    pos = 0
    s_len = len(s)

    while pos < s_len:
        m = _UNQUOTE_RE.search(s, pos=pos)
        if m is None:
            result.append(s[pos:])
            pos = s_len
        else:
            start = m.start()
            result.append(s[pos:start])
            pos = start + 4
            a = array.array("B")
            a.append(int(s[start + 1 : pos], 8))
            try:
                # Python >= 3.2
                result.append(a.tobytes())
            except AttributeError:
                result.append(a.tostring())

    return b"".join(result)


def dump_xattrs(pathnames, file_out):
    """Dump the xattr data for |pathnames| to |file_out|"""
    # NOTE: Always quote backslashes, in order to ensure that they are
    # not interpreted as quotes when they are processed by unquote.
    quote_chars = b"\n\r\\\\"

    for pathname in pathnames:
        attrs = xattr.list(pathname)
        if not attrs:
            continue

        file_out.write(b"# file: %s\n" % quote(pathname, quote_chars))
        for attr in attrs:
            attr = unicode_encode(attr)
            value = xattr.get(pathname, attr)
            file_out.write(
                b'%s="%s"\n'
                % (quote(attr, b"=" + quote_chars), quote(value, b'\0"' + quote_chars))
            )


def restore_xattrs(file_in):
    """Read |file_in| and restore xattrs content from it

    This expects textual data in the format written by dump_xattrs.
    """
    pathname = None
    for i, line in enumerate(file_in):
        if line.startswith(b"# file: "):
            pathname = unquote(line.rstrip(b"\n")[8:])
        else:
            parts = line.split(b"=", 1)
            if len(parts) == 2:
                if pathname is None:
                    raise ValueError("line %d: missing pathname" % (i + 1,))
                attr = unquote(parts[0])
                # strip trailing newline and quotes
                value = unquote(parts[1].rstrip(b"\n")[1:-1])
                xattr.set(pathname, attr, value)
            elif line.strip():
                raise ValueError("line %d: malformed entry" % (i + 1,))


def main(argv):

    parser = argparse.ArgumentParser(description=doc)
    parser.add_argument("paths", nargs="*", default=[])

    actions = parser.add_argument_group("Actions")
    actions.add_argument(
        "--dump",
        action="store_true",
        help="Dump the values of all extended "
        "attributes associated with paths "
        "passed as arguments or null-separated "
        "paths read from stdin.",
    )
    actions.add_argument(
        "--restore",
        action="store_true",
        help="Restore extended attributes using " "a dump read from stdin.",
    )

    options = parser.parse_args(argv)

    file_in = sys.stdin.buffer.raw

    if options.dump:
        if options.paths:
            options.paths = [unicode_encode(x) for x in options.paths]
        else:
            options.paths = [x for x in file_in.read().split(b"\0") if x]
        file_out = sys.stdout.buffer
        dump_xattrs(options.paths, file_out)

    elif options.restore:
        restore_xattrs(file_in)

    else:
        parser.error("missing action!")

    return os.EX_OK


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright 2022 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import tempfile

from pathlib import Path
from portage import Manifest
from portage.tests import TestCase


class ManifestTestCase(TestCase):
    def test_simple_addFile(self):
        tempdir = Path(tempfile.mkdtemp()) / "app-portage" / "diffball"
        manifest = Manifest(str(tempdir), required_hashes=["SHA512", "BLAKE2B"])

        (tempdir / "files").mkdir(parents=True)
        (tempdir / "files" / "test.patch").write_text(
            "Fix the diffball foobar functionality.\n"
        )

        # Nothing should be in the Manifest yet
        with self.assertRaises(KeyError):
            manifest.getFileData("AUX", "test.patch", "SHA512")

        manifest.addFile("AUX", "files/test.patch")

        self.assertEqual(len(manifest.fhashdict["AUX"].keys()), 1)
        self.assertEqual(
            manifest.getFileData("AUX", "test.patch", "SHA512"),
            "e30d069dcf284cbcb2d5685f03ca362469026b469dec4f8655d0c9a2bf317f5d9f68f61855ea403f4959bc0b9c003ae824fb9d6ab2472a739950623523af9da9",
        )



# Copyright 2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotConflictUpdateVirtTestCase(TestCase):
    def testSlotConflictUpdateVirt(self):

        ebuilds = {
            "dev-db/mysql-connector-c-6.1.11-r2": {"EAPI": "7", "SLOT": "0/18"},
            "dev-db/mysql-connector-c-8.0.17-r3": {"EAPI": "7", "SLOT": "0/21"},
            "virtual/libmysqlclient-18-r1": {
                "EAPI": "7",
                "SLOT": "0/18",
                "RDEPEND": "dev-db/mysql-connector-c:0/18",
            },
            "virtual/libmysqlclient-21": {
                "EAPI": "7",
                "SLOT": "0/21",
                "RDEPEND": "dev-db/mysql-connector-c:0/21",
            },
            "dev-perl/DBD-mysql-4.44.0": {
                "EAPI": "7",
                "RDEPEND": "virtual/libmysqlclient:=",
            },
        }

        installed = {
            "dev-db/mysql-connector-c-6.1.11-r2": {"EAPI": "7", "SLOT": "0/18"},
            "virtual/libmysqlclient-18-r1": {
                "EAPI": "7",
                "SLOT": "0/18",
                "RDEPEND": "dev-db/mysql-connector-c:0/18",
            },
            "dev-perl/DBD-mysql-4.44.0": {
                "EAPI": "7",
                "RDEPEND": "virtual/libmysqlclient:0/18=",
            },
        }

        world = ["dev-db/mysql-connector-c", "dev-perl/DBD-mysql"]

        test_cases = (
            # In order to avoid missed updates for bug 692746, consider
            # masking a package matched by all parent atoms.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[
                    "dev-db/mysql-connector-c-8.0.17-r3",
                    "virtual/libmysqlclient-21",
                    "dev-perl/DBD-mysql-4.44.0",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2010-2018 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["ExtractKernelVersion"]

import logging

from portage import os, _encodings, _unicode_encode
from portage.env.loaders import KeyValuePairFileLoader
from portage.util import grabfile, shlex_split, writemsg_level


def ExtractKernelVersion(base_dir):
    """
    Try to figure out what kernel version we are running
    @param base_dir: Path to sources (usually /usr/src/linux)
    @type base_dir: string
    @rtype: tuple( version[string], error[string])
    @return:
    1. tuple( version[string], error[string])
    Either version or error is populated (but never both)

    """
    lines = []
    pathname = os.path.join(base_dir, "Makefile")
    try:
        f = open(
            _unicode_encode(pathname, encoding=_encodings["fs"], errors="strict"),
            encoding=_encodings["content"],
            errors="replace",
        )
    except OSError as details:
        return (None, str(details))
    except OSError as details:
        return (None, str(details))

    try:
        for i in range(4):
            lines.append(f.readline())
    except OSError as details:
        return (None, str(details))
    except OSError as details:
        return (None, str(details))
    finally:
        f.close()

    lines = [l.strip() for l in lines]

    version = ""

    # XXX: The following code relies on the ordering of vars within the Makefile
    for line in lines:
        # split on the '=' then remove annoying whitespace
        items = line.split("=")
        items = [i.strip() for i in items]
        if items[0] == "VERSION" or items[0] == "PATCHLEVEL":
            version += items[1]
            version += "."
        elif items[0] == "SUBLEVEL":
            version += items[1]
        elif items[0] == "EXTRAVERSION" and items[-1] != items[0]:
            version += items[1]

    # Grab a list of files named localversion* and sort them
    localversions = os.listdir(base_dir)
    for x in range(len(localversions) - 1, -1, -1):
        if localversions[x][:12] != "localversion":
            del localversions[x]
    localversions.sort()

    # Append the contents of each to the version string, stripping ALL whitespace
    for lv in localversions:
        version += "".join(" ".join(grabfile(base_dir + "/" + lv)).split())

    # Check the .config for a CONFIG_LOCALVERSION and append that too, also stripping whitespace
    loader = KeyValuePairFileLoader(os.path.join(base_dir, ".config"), None)
    kernelconfig, loader_errors = loader.load()
    if loader_errors:
        for file_path, file_errors in loader_errors.items():
            for error_str in file_errors:
                writemsg_level(
                    "{}: {}\n".format(file_path, error_str),
                    level=logging.ERROR,
                    noiselevel=-1,
                )

    if kernelconfig and "CONFIG_LOCALVERSION" in kernelconfig:
        version += "".join(shlex_split(kernelconfig["CONFIG_LOCALVERSION"]))

    return (version, None)



# Copyright 2005-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

doc = """Check and fix problems in the resume and/or resume_backup files."""
__doc__ = doc


module_spec = {
    "name": "resume",
    "description": doc,
    "provides": {
        "module1": {
            "name": "cleanresume",
            "sourcefile": "resume",
            "class": "CleanResume",
            "description": "Discard emerge --resume merge lists",
            "functions": ["check", "fix"],
            "func_desc": {},
        }
    },
}



# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import subprocess

from portage.process import find_binary
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.futures import asyncio
from portage.util.futures._asyncio import create_subprocess_exec
from portage.util.futures.unix_events import DefaultEventLoopPolicy


class SubprocessExecTestCase(TestCase):
    def _run_test(self, test):
        initial_policy = asyncio.get_event_loop_policy()
        if not isinstance(initial_policy, DefaultEventLoopPolicy):
            asyncio.set_event_loop_policy(DefaultEventLoopPolicy())

        loop = asyncio._wrap_loop()
        try:
            test(loop)
        finally:
            asyncio.set_event_loop_policy(initial_policy)
            if loop not in (None, global_event_loop()):
                loop.close()
                self.assertFalse(global_event_loop().is_closed())

    def testEcho(self):
        args_tuple = (b"hello", b"world")
        echo_binary = find_binary("echo")
        self.assertNotEqual(echo_binary, None)
        echo_binary = echo_binary.encode()

        def test(loop):
            async def test_coroutine():

                proc = await create_subprocess_exec(
                    echo_binary,
                    *args_tuple,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT
                )

                out, err = await proc.communicate()
                self.assertEqual(tuple(out.split()), args_tuple)
                self.assertEqual(proc.returncode, os.EX_OK)

                proc = await create_subprocess_exec(
                    "bash",
                    "-c",
                    "echo foo; echo bar 1>&2;",
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

                out, err = await proc.communicate()
                self.assertEqual(out, b"foo\n")
                self.assertEqual(err, b"bar\n")
                self.assertEqual(proc.returncode, os.EX_OK)

                proc = await create_subprocess_exec(
                    "bash",
                    "-c",
                    "echo foo; echo bar 1>&2;",
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                )

                out, err = await proc.communicate()
                self.assertEqual(out, b"foo\nbar\n")
                self.assertEqual(err, None)
                self.assertEqual(proc.returncode, os.EX_OK)

                return "success"

            self.assertEqual("success", loop.run_until_complete(test_coroutine()))

        self._run_test(test)

    def testCat(self):
        stdin_data = b"hello world"
        cat_binary = find_binary("cat")
        self.assertNotEqual(cat_binary, None)
        cat_binary = cat_binary.encode()

        def test(loop):
            proc = loop.run_until_complete(
                create_subprocess_exec(
                    cat_binary, stdin=subprocess.PIPE, stdout=subprocess.PIPE, loop=loop
                )
            )

            out, err = loop.run_until_complete(proc.communicate(input=stdin_data))

            self.assertEqual(loop.run_until_complete(proc.wait()), os.EX_OK)
            self.assertEqual(out, stdin_data)

        self._run_test(test)

    def testPipe(self):
        stdin_data = b"hello world"
        cat_binary = find_binary("cat")
        self.assertNotEqual(cat_binary, None)
        cat_binary = cat_binary.encode()

        echo_binary = find_binary("echo")
        self.assertNotEqual(echo_binary, None)
        echo_binary = echo_binary.encode()

        def test(loop):

            pr, pw = os.pipe()

            cat_proc = loop.run_until_complete(
                create_subprocess_exec(
                    cat_binary, stdin=pr, stdout=subprocess.PIPE, loop=loop
                )
            )

            echo_proc = loop.run_until_complete(
                create_subprocess_exec(
                    echo_binary, b"-n", stdin_data, stdout=pw, loop=loop
                )
            )

            os.close(pr)
            os.close(pw)

            out, err = loop.run_until_complete(cat_proc.communicate())

            self.assertEqual(loop.run_until_complete(cat_proc.wait()), os.EX_OK)
            self.assertEqual(loop.run_until_complete(echo_proc.wait()), os.EX_OK)
            self.assertEqual(out, stdin_data)

        self._run_test(test)

    def testReadTransport(self):
        """
        Test asyncio.create_subprocess_exec(stdout=subprocess.PIPE) which
        requires an AbstractEventLoop.connect_read_pipe implementation
        (and a ReadTransport implementation for it to return).
        """

        args_tuple = (b"hello", b"world")
        echo_binary = find_binary("echo")
        self.assertNotEqual(echo_binary, None)
        echo_binary = echo_binary.encode()

        def test(loop):
            with open(os.devnull, "rb", 0) as devnull:
                proc = loop.run_until_complete(
                    create_subprocess_exec(
                        echo_binary,
                        *args_tuple,
                        stdin=devnull,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.STDOUT,
                        loop=loop
                    )
                )

            self.assertEqual(
                tuple(loop.run_until_complete(proc.stdout.read()).split()), args_tuple
            )
            self.assertEqual(loop.run_until_complete(proc.wait()), os.EX_OK)

        self._run_test(test)

    def testWriteTransport(self):
        """
        Test asyncio.create_subprocess_exec(stdin=subprocess.PIPE) which
        requires an AbstractEventLoop.connect_write_pipe implementation
        (and a WriteTransport implementation for it to return).
        """

        stdin_data = b"hello world"
        cat_binary = find_binary("cat")
        self.assertNotEqual(cat_binary, None)
        cat_binary = cat_binary.encode()

        def test(loop):
            proc = loop.run_until_complete(
                create_subprocess_exec(
                    cat_binary,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    loop=loop,
                )
            )

            # This buffers data when necessary to avoid blocking.
            proc.stdin.write(stdin_data)
            # Any buffered data is written asynchronously after the
            # close method is called.
            proc.stdin.close()

            self.assertEqual(loop.run_until_complete(proc.stdout.read()), stdin_data)
            self.assertEqual(loop.run_until_complete(proc.wait()), os.EX_OK)

        self._run_test(test)



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

doc = """WebRSync plug-in module for portage.
Performs a http download of a portage snapshot, verifies and
unpacks it to the repo location."""
__doc__ = doc[:]


import os

from portage.sync.config_checks import CheckSyncConfig


DEFAULT_CLASS = "WebRsync"
AVAILABLE_CLASSES = ["WebRsync", "PyWebsync"]
options = {"1": "WebRsync", "2": "PyWebsync"}


config_class = DEFAULT_CLASS
try:
    test_param = os.environ["TESTIT"]
    if test_param in options:
        config_class = options[test_param]
except KeyError:
    pass


module_spec = {
    "name": "webrsync",
    "description": doc,
    "provides": {
        "webrsync-module": {
            "name": "webrsync",
            "sourcefile": "webrsync",
            "class": config_class,
            "description": doc,
            "functions": ["sync", "new", "exists"],
            "func_desc": {
                "sync": "Performs an archived http download of the "
                + "repository, then unpacks it.  Optionally it performs a "
                + "gpg verification of the downloaded file(s)",
                "new": "Creates the new repository at the specified location",
                "exists": "Returns a boolean of whether the specified dir "
                + "exists and is a valid repository",
            },
            "validate_config": CheckSyncConfig,
            "module_specific_options": (
                "sync-webrsync-delta",
                "sync-webrsync-keep-snapshots",
                "sync-webrsync-verify-signature",
            ),
        },
    },
}



# Copyright 2018-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from concurrent.futures import Future, ThreadPoolExecutor
import contextlib

try:
    import threading
except ImportError:
    import dummy_threading as threading

import weakref
import time

import portage
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.backoff import RandomExponentialBackoff
from portage.util.futures import asyncio
from portage.util.futures.retry import retry
from portage.util.futures.executor.fork import ForkExecutor


class SucceedLaterException(Exception):
    pass


class SucceedLater:
    """
    A callable object that succeeds some duration of time has passed.
    """

    def __init__(self, duration):
        self._succeed_time = time.monotonic() + duration

    async def __call__(self):
        remaining = self._succeed_time - time.monotonic()
        if remaining > 0:
            await asyncio.sleep(remaining)
        return "success"


class SucceedNeverException(Exception):
    pass


class SucceedNever:
    """
    A callable object that never succeeds.
    """

    async def __call__(self):
        raise SucceedNeverException("expected failure")


class HangForever:
    """
    A callable object that sleeps forever.
    """

    async def __call__(self):
        while True:
            await asyncio.sleep(9)


class RetryTestCase(TestCase):
    @contextlib.contextmanager
    def _wrap_coroutine_func(self, coroutine_func):
        """
        Derived classes may override this method in order to implement
        alternative forms of execution.
        """
        yield coroutine_func

    def testSucceedLater(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedLater(1)) as func_coroutine:
            decorator = retry(
                try_max=9999,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            result = loop.run_until_complete(decorated_func())
            self.assertEqual(result, "success")

    def testSucceedNever(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedNever()) as func_coroutine:
            decorator = retry(
                try_max=4,
                try_timeout=None,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(
                isinstance(done.pop().exception().__cause__, SucceedNeverException)
            )

    def testSucceedNeverReraise(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedNever()) as func_coroutine:
            decorator = retry(
                reraise=True,
                try_max=4,
                try_timeout=None,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(isinstance(done.pop().exception(), SucceedNeverException))

    def testHangForever(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(HangForever()) as func_coroutine:
            decorator = retry(
                try_max=2,
                try_timeout=0.1,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(
                isinstance(done.pop().exception().__cause__, asyncio.TimeoutError)
            )

    def testHangForeverReraise(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(HangForever()) as func_coroutine:
            decorator = retry(
                reraise=True,
                try_max=2,
                try_timeout=0.1,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(isinstance(done.pop().exception(), asyncio.TimeoutError))

    def testCancelRetry(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedNever()) as func_coroutine:
            decorator = retry(
                try_timeout=0.1,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            future = decorated_func()
            loop.call_later(0.3, future.cancel)
            done, pending = loop.run_until_complete(asyncio.wait([future], loop=loop))
            self.assertEqual(len(done), 1)
            self.assertTrue(done.pop().cancelled())

    def testOverallTimeoutWithException(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedNever()) as func_coroutine:
            decorator = retry(
                try_timeout=0.1,
                overall_timeout=0.3,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            cause = done.pop().exception().__cause__
            self.assertTrue(
                isinstance(
                    cause,
                    (asyncio.TimeoutError, SucceedNeverException),
                ),
                msg=f"Cause was {cause.__class__.__name__}",
            )

    def testOverallTimeoutWithTimeoutError(self):
        loop = global_event_loop()
        # results in TimeoutError because it hangs forever
        with self._wrap_coroutine_func(HangForever()) as func_coroutine:
            decorator = retry(
                try_timeout=0.1,
                overall_timeout=0.3,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(
                isinstance(done.pop().exception().__cause__, asyncio.TimeoutError)
            )


class RetryForkExecutorTestCase(RetryTestCase):
    """
    Wrap each coroutine function with AbstractEventLoop.run_in_executor,
    in order to test the event loop's default executor. The executor
    may use either a thread or a subprocess, and either case is
    automatically detected and handled.
    """

    def __init__(self, *pargs, **kwargs):
        super().__init__(*pargs, **kwargs)
        self._executor = None

    def _setUpExecutor(self):
        self._executor = ForkExecutor()

    def _tearDownExecutor(self):
        if self._executor is not None:
            self._executor.shutdown(wait=True)
            self._executor = None

    def setUp(self):
        self._setUpExecutor()

    def tearDown(self):
        self._tearDownExecutor()

    @contextlib.contextmanager
    def _wrap_coroutine_func(self, coroutine_func):
        parent_loop = global_event_loop()
        parent_pid = portage.getpid()
        pending = weakref.WeakValueDictionary()

        # Since ThreadPoolExecutor does not propagate cancellation of a
        # parent_future to the underlying coroutine, use kill_switch to
        # propagate task cancellation to wrapper, so that HangForever's
        # thread returns when retry eventually cancels parent_future.
        def wrapper(kill_switch):
            if portage.getpid() == parent_pid:
                # thread in main process
                def done_callback(result):
                    result.cancelled() or result.exception() or result.result()
                    kill_switch.set()

                def start_coroutine(future):
                    result = asyncio.ensure_future(coroutine_func(), loop=parent_loop)
                    pending[id(result)] = result
                    result.add_done_callback(done_callback)
                    future.set_result(result)

                future = Future()
                parent_loop.call_soon_threadsafe(start_coroutine, future)
                kill_switch.wait()
                if not future.done():
                    future.cancel()
                    raise asyncio.CancelledError
                elif not future.result().done():
                    future.result().cancel()
                    raise asyncio.CancelledError
                else:
                    return future.result().result()

            # child process
            loop = global_event_loop()
            try:
                return loop.run_until_complete(coroutine_func())
            finally:
                loop.close()

        def execute_wrapper():
            kill_switch = threading.Event()
            parent_future = asyncio.ensure_future(
                parent_loop.run_in_executor(self._executor, wrapper, kill_switch),
                loop=parent_loop,
            )

            def kill_callback(parent_future):
                if not kill_switch.is_set():
                    kill_switch.set()

            parent_future.add_done_callback(kill_callback)
            return parent_future

        try:
            yield execute_wrapper
        finally:
            while True:
                try:
                    _, future = pending.popitem()
                except KeyError:
                    break
                try:
                    parent_loop.run_until_complete(future)
                except (Exception, asyncio.CancelledError):
                    pass
                future.cancelled() or future.exception() or future.result()


class RetryThreadExecutorTestCase(RetryForkExecutorTestCase):
    def _setUpExecutor(self):
        self._executor = ThreadPoolExecutor(max_workers=1)



import signal


def run_main_scheduler(scheduler):
    """
    Start and run an AsyncScheduler (or compatible object), and handle
    SIGINT or SIGTERM by calling its terminate() method and waiting
    for it to clean up after itself. If SIGINT or SIGTERM is received,
    return signum, else return None. Any previous SIGINT or SIGTERM
    signal handlers are automatically saved and restored before
    returning.
    """

    received_signal = []

    def sighandler(signum, frame):
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        received_signal.append(signum)
        scheduler.terminate()

    earlier_sigint_handler = signal.signal(signal.SIGINT, sighandler)
    earlier_sigterm_handler = signal.signal(signal.SIGTERM, sighandler)

    try:
        scheduler.start()
        scheduler.wait()
    finally:
        # Restore previous handlers
        if earlier_sigint_handler is not None:
            signal.signal(signal.SIGINT, earlier_sigint_handler)
        else:
            signal.signal(signal.SIGINT, signal.SIG_DFL)
        if earlier_sigterm_handler is not None:
            signal.signal(signal.SIGTERM, earlier_sigterm_handler)
        else:
            signal.signal(signal.SIGTERM, signal.SIG_DFL)

    if received_signal:
        return received_signal[0]
    return None



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import locale
import logging
import time

from portage import os, _unicode_decode
from portage.exception import PortageException
from portage.localization import _
from portage.output import EOutput
from portage.util import grabfile, writemsg_level


def have_english_locale():
    lang, enc = locale.getdefaultlocale()
    if lang is not None:
        lang = lang.lower()
        lang = lang.split("_", 1)[0]
    return lang is None or lang in ("c", "en")


def whenago(seconds):
    sec = int(seconds)
    mins = 0
    days = 0
    hrs = 0
    years = 0
    out = []

    if sec > 60:
        mins = sec // 60
        sec = sec % 60
    if mins > 60:
        hrs = mins // 60
        mins = mins % 60
    if hrs > 24:
        days = hrs // 24
        hrs = hrs % 24
    if days > 365:
        years = days // 365
        days = days % 365

    if years:
        out.append("%dy " % years)
    if days:
        out.append("%dd " % days)
    if hrs:
        out.append("%dh " % hrs)
    if mins:
        out.append("%dm " % mins)
    if sec:
        out.append("%ds " % sec)

    return "".join(out).strip()


def old_tree_timestamp_warn(portdir, settings):
    unixtime = time.time()
    default_warnsync = 30

    timestamp_file = os.path.join(portdir, "metadata/timestamp.x")
    try:
        lastsync = grabfile(timestamp_file)
    except PortageException:
        return False

    if not lastsync:
        return False

    lastsync = lastsync[0].split()
    if not lastsync:
        return False

    try:
        lastsync = int(lastsync[0])
    except ValueError:
        return False

    var_name = "PORTAGE_SYNC_STALE"
    try:
        warnsync = float(settings.get(var_name, default_warnsync))
    except ValueError:
        writemsg_level(
            "!!! {} contains non-numeric value: {}\n".format(
                var_name, settings[var_name]
            ),
            level=logging.ERROR,
            noiselevel=-1,
        )
        return False

    if warnsync <= 0:
        return False

    if (unixtime - 86400 * warnsync) > lastsync:
        out = EOutput()
        if have_english_locale():
            out.ewarn("Last emerge --sync was %s ago." % whenago(unixtime - lastsync))
        else:
            out.ewarn(
                _("Last emerge --sync was %s.")
                % _unicode_decode(time.strftime("%c", time.localtime(lastsync)))
            )
        return True
    return False



# tests/__init__.py -- Portage Unit Test functionality
# Copyright 2006-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import argparse
import sys
import time
import unittest
from pathlib import Path

from unittest.runner import TextTestResult as _TextTestResult

import portage
from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage.output import colorize
from portage.proxy.objectproxy import ObjectProxy


# This remains constant when the real value is a mock.
EPREFIX_ORIG = portage.const.EPREFIX


class lazy_value(ObjectProxy):
    __slots__ = ("_func",)

    def __init__(self, func):
        ObjectProxy.__init__(self)
        object.__setattr__(self, "_func", func)

    def _get_target(self):
        return object.__getattribute__(self, "_func")()


@lazy_value
def cnf_path():
    if portage._not_installed:
        return os.path.join(portage.const.PORTAGE_BASE_PATH, "cnf")
    return os.path.join(
        EPREFIX_ORIG or "/", portage.const.GLOBAL_CONFIG_PATH.lstrip(os.sep)
    )


@lazy_value
def cnf_etc_path():
    if portage._not_installed:
        return str(cnf_path)
    return os.path.join(EPREFIX_ORIG or "/", "etc")


@lazy_value
def cnf_bindir():
    if portage._not_installed:
        return portage.const.PORTAGE_BIN_PATH
    return os.path.join(portage.const.EPREFIX or "/", "usr", "bin")


@lazy_value
def cnf_sbindir():
    if portage._not_installed:
        return str(cnf_bindir)
    return os.path.join(portage.const.EPREFIX or "/", "usr", "sbin")


def main():
    suite = unittest.TestSuite()
    basedir = Path(__file__).resolve().parent

    argv0 = Path(sys.argv[0])

    usage = "usage: %s [options] [tests to run]" % argv0.name
    parser = argparse.ArgumentParser(usage=usage)
    parser.add_argument(
        "-l", "--list", help="list all tests", action="store_true", dest="list_tests"
    )
    parser.add_argument("tests", nargs="*", type=Path)
    options = parser.parse_args(args=sys.argv)

    if (
        os.environ.get("NOCOLOR") in ("yes", "true")
        or os.environ.get("TERM") == "dumb"
        or not sys.stdout.isatty()
    ):
        portage.output.nocolor()

    if options.list_tests:
        testdir = argv0.parent
        for mydir in getTestDirs(basedir):
            testsubdir = mydir.name
            for name in getTestNames(mydir):
                print("{}/{}/{}.py".format(testdir, testsubdir, name))
        return os.EX_OK

    if len(options.tests) > 1:
        suite.addTests(getTestFromCommandLine(options.tests[1:], basedir))
    else:
        for mydir in getTestDirs(basedir):
            suite.addTests(getTests(mydir, basedir))

    result = TextTestRunner(verbosity=2).run(suite)
    if not result.wasSuccessful():
        return 1
    return os.EX_OK


def my_import(name):
    mod = __import__(name)
    components = name.split(".")
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def getTestFromCommandLine(args, base_path):
    result = []
    for arg in args:
        realpath = arg.resolve()
        path = realpath.parent
        f = realpath.relative_to(path)

        if not f.name.startswith("test") or not f.suffix == ".py":
            raise Exception("Invalid argument: '%s'" % arg)

        mymodule = f.stem
        result.extend(getTestsFromFiles(path, base_path, [mymodule]))
    return result


def getTestDirs(base_path):
    TEST_FILE = "__test__.py"
    testDirs = []

    # the os.walk help mentions relative paths as being quirky
    # I was tired of adding dirs to the list, so now we add __test__.py
    # to each dir we want tested.
    for testFile in base_path.rglob(TEST_FILE):
        testDirs.append(testFile.parent)

    testDirs.sort()
    return testDirs


def getTestNames(path):
    files = path.glob("*")
    files = [f.stem for f in files if f.name.startswith("test") and f.suffix == ".py"]
    files.sort()
    return files


def getTestsFromFiles(path, base_path, files):
    parent_path = path.relative_to(base_path)
    parent_module = ".".join(("portage", "tests") + parent_path.parts)
    result = []
    for mymodule in files:
        # Make the trailing / a . for module importing
        modname = ".".join((parent_module, mymodule))
        mod = my_import(modname)
        result.append(unittest.TestLoader().loadTestsFromModule(mod))
    return result


def getTests(path, base_path):
    """

    path is the path to a given subdir ( 'portage/' for example)
    This does a simple filter on files in that dir to give us modules
    to import

    """
    return getTestsFromFiles(path, base_path, getTestNames(path))


class TextTestResult(_TextTestResult):
    """
    We need a subclass of unittest.runner.TextTestResult to handle tests with TODO

    This just adds an addTodo method that can be used to add tests
    that are marked TODO; these can be displayed later
    by the test runner.
    """

    def __init__(self, stream, descriptions, verbosity):
        super().__init__(stream, descriptions, verbosity)
        self.todoed = []
        self.portage_skipped = []

    def addSuccess(self, test):
        super(_TextTestResult, self).addSuccess(test)
        if self.showAll:
            self.stream.writeln(colorize("GOOD", "ok"))
        elif self.dots:
            self.stream.write(colorize("GOOD", "."))
            self.stream.flush()

    def addError(self, test, err):
        super(_TextTestResult, self).addError(test, err)
        if self.showAll:
            self.stream.writeln(colorize("BAD", "ERROR"))
        elif self.dots:
            self.stream.write(colorize("HILITE", "E"))
            self.stream.flush()

    def addFailure(self, test, err):
        super(_TextTestResult, self).addFailure(test, err)
        if self.showAll:
            self.stream.writeln(colorize("BAD", "FAIL"))
        elif self.dots:
            self.stream.write(colorize("BAD", "F"))
            self.stream.flush()

    def addTodo(self, test, info):
        self.todoed.append((test, info))
        if self.showAll:
            self.stream.writeln(colorize("BRACKET", "TODO"))
        elif self.dots:
            self.stream.write(colorize("BRACKET", "."))

    def addPortageSkip(self, test, info):
        self.portage_skipped.append((test, info))
        if self.showAll:
            self.stream.writeln(colorize("WARN", "SKIP"))
        elif self.dots:
            self.stream.write(colorize("WARN", "."))

    def printErrors(self):
        if self.dots or self.showAll:
            self.stream.writeln()
            self.printErrorList("ERROR", self.errors)
            self.printErrorList("FAIL", self.failures)
            self.printErrorList("TODO", self.todoed)
            self.printErrorList("SKIP", self.portage_skipped)


class TestCase(unittest.TestCase):
    """
    We need a way to mark a unit test as "ok to fail"
    This way someone can add a broken test and mark it as failed
    and then fix the code later.  This may not be a great approach
    (broken code!!??!11oneone) but it does happen at times.
    """

    def __init__(self, *pargs, **kwargs):
        unittest.TestCase.__init__(self, *pargs, **kwargs)
        self.todo = False
        self.portage_skip = None
        self.cnf_path = cnf_path
        self.cnf_etc_path = cnf_etc_path
        self.bindir = cnf_bindir
        self.sbindir = cnf_sbindir

    def defaultTestResult(self):
        return TextTestResult()

    def run(self, result=None):
        if result is None:
            result = self.defaultTestResult()
        result.startTest(self)
        testMethod = getattr(self, self._testMethodName)
        try:
            ok = False
            try:
                try:
                    self.setUp()
                except KeyboardInterrupt:
                    raise
                except unittest.SkipTest:
                    raise
                except Exception:
                    result.addError(self, sys.exc_info())
                    return

                testMethod()
                ok = True
            except unittest.SkipTest as e:
                result.addPortageSkip(self, "{}: SKIP: {}".format(testMethod, str(e)))
            except self.failureException:
                if self.portage_skip is not None:
                    if self.portage_skip is True:
                        result.addPortageSkip(self, "%s: SKIP" % testMethod)
                    else:
                        result.addPortageSkip(
                            self, "{}: SKIP: {}".format(testMethod, self.portage_skip)
                        )
                elif self.todo:
                    result.addTodo(self, "%s: TODO" % testMethod)
                else:
                    result.addFailure(self, sys.exc_info())
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                result.addError(self, sys.exc_info())

            try:
                self.tearDown()
            except SystemExit:
                raise
            except KeyboardInterrupt:
                raise
            except:
                result.addError(self, sys.exc_info())
                ok = False
            if ok:
                result.addSuccess(self)
        finally:
            result.stopTest(self)

    def assertRaisesMsg(self, msg, excClass, callableObj, *args, **kwargs):
        """Fail unless an exception of class excClass is thrown
        by callableObj when invoked with arguments args and keyword
        arguments kwargs. If a different type of exception is
        thrown, it will not be caught, and the test case will be
        deemed to have suffered an error, exactly as for an
        unexpected exception.
        """
        try:
            callableObj(*args, **kwargs)
        except excClass:
            return
        else:
            if hasattr(excClass, "__name__"):
                excName = excClass.__name__
            else:
                excName = str(excClass)
            raise self.failureException("{} not raised: {}".format(excName, msg))

    def assertNotExists(self, path):
        """Make sure |path| does not exist"""
        path = Path(path)
        if path.exists():
            raise self.failureException("path exists when it should not: %s" % path)


class TextTestRunner(unittest.TextTestRunner):
    """
    We subclass unittest.TextTestRunner to output SKIP for tests that fail but are skippable
    """

    def _makeResult(self):
        return TextTestResult(self.stream, self.descriptions, self.verbosity)

    def run(self, test):
        """
        Run the given test case or test suite.
        """
        result = self._makeResult()
        startTime = time.time()
        test(result)
        stopTime = time.time()
        timeTaken = stopTime - startTime
        result.printErrors()
        self.stream.writeln(result.separator2)
        run = result.testsRun
        self.stream.writeln(
            "Ran %d test%s in %.3fs" % (run, run != 1 and "s" or "", timeTaken)
        )
        self.stream.writeln()
        if not result.wasSuccessful():
            self.stream.write(colorize("BAD", "FAILED") + " (")
            failed = len(result.failures)
            errored = len(result.errors)
            if failed:
                self.stream.write("failures=%d" % failed)
            if errored:
                if failed:
                    self.stream.write(", ")
                self.stream.write("errors=%d" % errored)
            self.stream.writeln(")")
        else:
            self.stream.writeln(colorize("GOOD", "OK"))
        return result


test_cps = ["sys-apps/portage", "virtual/portage"]
test_versions = ["1.0", "1.0-r1", "2.3_p4", "1.0_alpha57"]
test_slots = [None, "1", "gentoo-sources-2.6.17", "spankywashere"]
test_usedeps = ["foo", "-bar", ("foo", "bar"), ("foo", "-bar"), ("foo?", "!bar?")]



# Copyright 2013-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import itertools
import logging
import stat

from portage import os
from portage.package.ebuild.fetch import DistfileName
from .DeletionTask import DeletionTask

logger = logging.getLogger(__name__)


class DeletionIterator:
    def __init__(self, config):
        self._config = config

    def __iter__(self):
        distdir = self._config.options.distfiles
        file_owners = self._config.file_owners
        whitelist = self._config.whitelist
        distfiles_local = self._config.options.distfiles_local
        deletion_db = self._config.deletion_db
        deletion_delay = self._config.options.deletion_delay
        start_time = self._config.start_time
        distfiles_set = set()
        distfiles_set.update(
            (
                filename
                if isinstance(filename, DistfileName)
                else DistfileName(filename)
                for filename in itertools.chain.from_iterable(
                    layout.get_filenames(distdir) for layout in self._config.layouts
                )
            )
            if self._config.content_db is None
            else itertools.chain.from_iterable(
                self._config.content_db.get_filenames_translate(filename)
                for filename in itertools.chain.from_iterable(
                    layout.get_filenames(distdir) for layout in self._config.layouts
                )
            )
        )
        for filename in distfiles_set:
            # require at least one successful stat()
            exceptions = []
            for layout in reversed(self._config.layouts):
                path = os.path.join(distdir, layout.get_path(filename))
                try:
                    st = os.stat(path)
                except OSError as e:
                    # is it a dangling symlink?
                    try:
                        if os.path.islink(path):
                            os.unlink(path)
                    except OSError as e:
                        exceptions.append(e)
                else:
                    if stat.S_ISREG(st.st_mode):
                        break
            else:
                if exceptions:
                    logger.error(
                        "stat failed on '%s' in distfiles: %s\n"
                        % (filename, "; ".join(str(x) for x in exceptions))
                    )
                continue

            if filename in file_owners:
                if deletion_db is not None:
                    try:
                        del deletion_db[filename]
                    except KeyError:
                        pass
            elif whitelist is not None and filename in whitelist:
                if deletion_db is not None:
                    try:
                        del deletion_db[filename]
                    except KeyError:
                        pass
            elif distfiles_local is not None and os.path.exists(
                os.path.join(distfiles_local, filename)
            ):
                if deletion_db is not None:
                    try:
                        del deletion_db[filename]
                    except KeyError:
                        pass
            else:
                self._config.scheduled_deletion_count += 1

                if deletion_db is None or deletion_delay is None:

                    yield DeletionTask(
                        background=True,
                        distfile=filename,
                        distfile_path=path,
                        config=self._config,
                    )

                else:
                    deletion_entry = deletion_db.get(filename)

                    if deletion_entry is None:
                        logger.debug("add '%s' to deletion db" % filename)
                        deletion_db[filename] = start_time

                    elif deletion_entry + deletion_delay <= start_time:

                        yield DeletionTask(
                            background=True,
                            distfile=filename,
                            distfile_path=path,
                            config=self._config,
                        )

        if deletion_db is not None:
            for filename in list(deletion_db):
                if filename not in distfiles_set:
                    try:
                        del deletion_db[filename]
                    except KeyError:
                        pass
                    else:
                        logger.debug("drop '%s' from deletion db" % filename)



# checksum.py -- core Portage functionality
# Copyright 1998-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
# pylint: disable=ungrouped-imports

import errno
import functools
import hashlib
import portage
import stat
import subprocess
import tempfile

from portage import _encodings, _unicode_decode, _unicode_encode
from portage import os
from portage.const import HASHING_BLOCKSIZE, PRELINK_BINARY
from portage.localization import _


# Summary of all available hashes and their implementations,
# most preferred first. Please keep this in sync with logic below.
# ================================================================
#
# MD5: hashlib
# SHA1: hashlib
# SHA256: hashlib
# SHA512: hashlib
# RMD160: hashlib, pycrypto, mhash
# WHIRLPOOL: hashlib, mhash, bundled
# BLAKE2B (512): hashlib (3.6+), pyblake2, pycrypto
# BLAKE2S (512): hashlib (3.6+), pyblake2, pycrypto
# SHA3_256: hashlib (3.6+), pysha3, pycrypto
# SHA3_512: hashlib (3.6+), pysha3, pycrypto


# Dict of all available hash functions
hashfunc_map = {}
hashorigin_map = {}


def _open_file(filename):
    try:
        return open(
            _unicode_encode(filename, encoding=_encodings["fs"], errors="strict"), "rb"
        )
    except OSError as e:
        func_call = f"open('{_unicode_decode(filename)}')"
        if e.errno == errno.EPERM:
            raise portage.exception.OperationNotPermitted(func_call)
        elif e.errno == errno.EACCES:
            raise portage.exception.PermissionDenied(func_call)
        elif e.errno == errno.ENOENT:
            raise portage.exception.FileNotFound(filename)
        else:
            raise


class _generate_hash_function:

    __slots__ = ("_hashobject",)

    def __init__(self, hashtype, hashobject, origin="unknown"):
        self._hashobject = hashobject
        hashfunc_map[hashtype] = self
        hashorigin_map[hashtype] = origin

    def checksum_str(self, data):
        """
        Obtain a checksum of a byte-string.

        @param data: Data to hash
        @type data: bytes
        @return: The hash of the data (hex-digest)
        """
        checksum = self._hashobject()
        checksum.update(data)
        return checksum.hexdigest()

    def checksum_file(self, filename):
        """
        Run a checksum against a file.

        @param filename: File to run the checksum against
        @type filename: String
        @return: The hash and size of the data
        """
        with _open_file(filename) as f:
            blocksize = HASHING_BLOCKSIZE
            size = 0
            checksum = self._hashobject()
            data = f.read(blocksize)
            while data:
                checksum.update(data)
                size = size + len(data)
                data = f.read(blocksize)

        return (checksum.hexdigest(), size)


# Define hash functions, try to use the best module available. Preferred
# modules should go first, latter ones should check if the hashes aren't
# already defined.


# Use hashlib from python-2.5 if available and prefer it over pycrypto and internal fallbacks.
# Need special handling for RMD160/WHIRLPOOL as they may not always be provided by hashlib.
_generate_hash_function("MD5", hashlib.md5, origin="hashlib")
_generate_hash_function("SHA1", hashlib.sha1, origin="hashlib")
_generate_hash_function("SHA256", hashlib.sha256, origin="hashlib")
_generate_hash_function("SHA512", hashlib.sha512, origin="hashlib")
for local_name, hash_name in (
    ("RMD160", "ripemd160"),
    ("WHIRLPOOL", "whirlpool"),
    # available since Python 3.6
    ("BLAKE2B", "blake2b"),
    ("BLAKE2S", "blake2s"),
    ("SHA3_256", "sha3_256"),
    ("SHA3_512", "sha3_512"),
):
    try:
        hashlib.new(hash_name)
    except ValueError:
        pass
    else:
        _generate_hash_function(
            local_name, functools.partial(hashlib.new, hash_name), origin="hashlib"
        )


# Support using pyblake2 as fallback for python<3.6
if "BLAKE2B" not in hashfunc_map or "BLAKE2S" not in hashfunc_map:
    try:
        import pyblake2

        _generate_hash_function("BLAKE2B", pyblake2.blake2b, origin="pyblake2")
        _generate_hash_function("BLAKE2S", pyblake2.blake2s, origin="pyblake2")
    except ImportError:
        pass


# Support using pysha3 as fallback for python<3.6
if "SHA3_256" not in hashfunc_map or "SHA3_512" not in hashfunc_map:
    try:
        import sha3

        _generate_hash_function("SHA3_256", sha3.sha3_256, origin="pysha3")
        _generate_hash_function("SHA3_512", sha3.sha3_512, origin="pysha3")
    except ImportError:
        pass


# Support pygcrypt as fallback using optimized routines from libgcrypt
# (GnuPG).
gcrypt_algos = frozenset(
    ("RMD160", "WHIRLPOOL", "SHA3_256", "SHA3_512", "STREEBOG256", "STREEBOG512")
)
# Note: currently disabled due to resource exhaustion bugs in pygcrypt.
# Please do not reenable until upstream has a fix.
# https://bugs.gentoo.org/615620
if False:
    # if gcrypt_algos.difference(hashfunc_map):
    try:
        import binascii
        import pygcrypt.hashcontext

        class GCryptHashWrapper:
            def __init__(self, algo):
                self._obj = pygcrypt.hashcontext.HashContext(algo=algo, secure=False)

            def update(self, data):
                self._obj.write(data)

            def hexdigest(self):
                return binascii.b2a_hex(self._obj.read()).decode()

        name_mapping = {
            "RMD160": "ripemd160",
            "WHIRLPOOL": "whirlpool",
            "SHA3_256": "sha3-256",
            "SHA3_512": "sha3-512",
            "STREEBOG256": "stribog256",
            "STREEBOG512": "stribog512",
        }

        for local_name, gcry_name in name_mapping.items():
            try:
                pygcrypt.hashcontext.HashContext(algo=gcry_name)
            except Exception:  # yes, it throws Exception...
                pass
            else:
                _generate_hash_function(
                    local_name,
                    functools.partial(GCryptHashWrapper, gcry_name),
                    origin="pygcrypt",
                )
    except ImportError:
        pass


# Use pycrypto when available, prefer it over the internal fallbacks
# Check for 'new' attributes, since they can be missing if the module
# is broken somehow.
if "RMD160" not in hashfunc_map:
    try:
        from Crypto.Hash import RIPEMD

        rmd160hash_ = getattr(RIPEMD, "new", None)
        if rmd160hash_ is not None:
            _generate_hash_function("RMD160", rmd160hash_, origin="pycrypto")
    except ImportError:
        pass

# The following hashes were added in pycryptodome (pycrypto fork)
if "BLAKE2B" not in hashfunc_map:
    try:
        from Crypto.Hash import BLAKE2b

        blake2bhash_ = getattr(BLAKE2b, "new", None)
        if blake2bhash_ is not None:
            _generate_hash_function(
                "BLAKE2B",
                functools.partial(blake2bhash_, digest_bytes=64),
                origin="pycrypto",
            )
    except ImportError:
        pass

if "BLAKE2S" not in hashfunc_map:
    try:
        from Crypto.Hash import BLAKE2s

        blake2shash_ = getattr(BLAKE2s, "new", None)
        if blake2shash_ is not None:
            _generate_hash_function(
                "BLAKE2S",
                functools.partial(blake2shash_, digest_bytes=32),
                origin="pycrypto",
            )
    except ImportError:
        pass

if "SHA3_256" not in hashfunc_map:
    try:
        from Crypto.Hash import SHA3_256

        sha3_256hash_ = getattr(SHA3_256, "new", None)
        if sha3_256hash_ is not None:
            _generate_hash_function("SHA3_256", sha3_256hash_, origin="pycrypto")
    except ImportError:
        pass

if "SHA3_512" not in hashfunc_map:
    try:
        from Crypto.Hash import SHA3_512

        sha3_512hash_ = getattr(SHA3_512, "new", None)
        if sha3_512hash_ is not None:
            _generate_hash_function("SHA3_512", sha3_512hash_, origin="pycrypto")
    except ImportError:
        pass


# Try to use mhash if available
# mhash causes GIL presently, so it gets less priority than hashlib and
# pycrypto. However, it might be the only accelerated implementation of
# WHIRLPOOL available.
if "RMD160" not in hashfunc_map or "WHIRLPOOL" not in hashfunc_map:
    try:
        import mhash

        for local_name, hash_name in (
            ("RMD160", "RIPEMD160"),
            ("WHIRLPOOL", "WHIRLPOOL"),
        ):
            if local_name not in hashfunc_map and hasattr(
                mhash, "MHASH_%s" % hash_name
            ):
                _generate_hash_function(
                    local_name,
                    functools.partial(
                        mhash.MHASH, getattr(mhash, "MHASH_%s" % hash_name)
                    ),
                    origin="mhash",
                )
    except ImportError:
        pass


# Support pygost as fallback streebog provider
# It's mostly provided as a reference implementation; it's pure Python,
# slow and reads all data to memory (i.e. doesn't hash on update()...)
if "STREEBOG256" not in hashfunc_map or "STREEBOG512" not in hashfunc_map:
    try:
        import pygost.gost34112012

        _generate_hash_function(
            "STREEBOG256",
            functools.partial(pygost.gost34112012.GOST34112012, digest_size=32),
            origin="pygost",
        )
        _generate_hash_function(
            "STREEBOG512",
            functools.partial(pygost.gost34112012.GOST34112012, digest_size=64),
            origin="pygost",
        )
    except ImportError:
        pass


_whirlpool_unaccelerated = False
if "WHIRLPOOL" not in hashfunc_map:
    # Bundled WHIRLPOOL implementation
    _whirlpool_unaccelerated = True
    from portage.util.whirlpool import new as _new_whirlpool

    _generate_hash_function("WHIRLPOOL", _new_whirlpool, origin="bundled")


# There is only one implementation for size
class SizeHash:
    def checksum_file(self, filename):
        size = os.stat(filename).st_size
        return (size, size)


hashfunc_map["size"] = SizeHash()

# cache all supported hash methods in a frozenset
hashfunc_keys = frozenset(hashfunc_map)

# end actual hash functions


prelink_capable = False
if os.path.exists(PRELINK_BINARY):
    cmd = [PRELINK_BINARY, "--version"]
    cmd = [_unicode_encode(x, encoding=_encodings["fs"], errors="strict") for x in cmd]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    proc.communicate()
    status = proc.wait()
    if os.WIFEXITED(status) and os.WEXITSTATUS(status) == os.EX_OK:
        prelink_capable = True
    del cmd, proc, status


def is_prelinkable_elf(filename):
    with _open_file(filename) as f:
        magic = f.read(17)
    return (
        len(magic) == 17
        and magic.startswith(b"\x7fELF")
        and magic[16:17] in (b"\x02", b"\x03")
    )  # 2=ET_EXEC, 3=ET_DYN


def perform_md5(x, calc_prelink=0):
    return perform_checksum(x, "MD5", calc_prelink)[0]


def _perform_md5_merge(x, **kwargs):
    return perform_md5(
        _unicode_encode(x, encoding=_encodings["merge"], errors="strict"), **kwargs
    )


def perform_all(x, calc_prelink=0):
    mydict = {k: perform_checksum(x, k, calc_prelink)[0] for k in hashfunc_keys}
    return mydict


def get_valid_checksum_keys():
    return hashfunc_keys


def get_hash_origin(hashtype):
    if hashtype not in hashfunc_keys:
        raise KeyError(hashtype)
    return hashorigin_map.get(hashtype, "unknown")


def _filter_unaccelarated_hashes(digests):
    """
    If multiple digests are available and some are unaccelerated,
    then return a new dict that omits the unaccelerated ones. This
    allows extreme performance problems like bug #425046 to be
    avoided whenever practical, especially for cases like stage
    builds where acceleration may not be available for some hashes
    due to minimization of dependencies.
    """
    if _whirlpool_unaccelerated and "WHIRLPOOL" in digests:
        verifiable_hash_types = set(digests).intersection(hashfunc_keys)
        verifiable_hash_types.discard("size")
        if len(verifiable_hash_types) > 1:
            digests = dict(digests)
            digests.pop("WHIRLPOOL")

    return digests


class _hash_filter:
    """
    Implements filtering for PORTAGE_CHECKSUM_FILTER.
    """

    __slots__ = (
        "transparent",
        "_tokens",
    )

    def __init__(self, filter_str):
        tokens = filter_str.upper().split()
        if not tokens or tokens[-1] == "*":
            del tokens[:]
        self.transparent = not tokens
        tokens.reverse()
        self._tokens = tuple(tokens)

    def __call__(self, hash_name):
        if self.transparent:
            return True
        matches = ("*", hash_name)
        for token in self._tokens:
            if token in matches:
                return True
            if token[:1] == "-" and token[1:] in matches:
                return False
        return False


def _apply_hash_filter(digests, hash_filter):
    """
    Return a new dict containing the filtered digests, or the same
    dict if no changes are necessary. This will always preserve at
    at least one digest, in order to ensure that they are not all
    discarded.
    @param digests: dictionary of digests
    @type digests: dict
    @param hash_filter: A callable that takes a single hash name
            argument, and returns True if the hash is to be used or
            False otherwise
    @type hash_filter: callable
    """

    verifiable_hash_types = set(digests).intersection(hashfunc_keys)
    verifiable_hash_types.discard("size")
    modified = False
    if len(verifiable_hash_types) > 1:
        for k in list(verifiable_hash_types):
            if not hash_filter(k):
                modified = True
                verifiable_hash_types.remove(k)
                if len(verifiable_hash_types) == 1:
                    break

    if modified:
        digests = {
            k: v
            for k, v in digests.items()
            if k == "size" or k in verifiable_hash_types
        }

    return digests


def verify_all(filename, mydict, calc_prelink=0, strict=0):
    """
    Verify all checksums against a file.

    @param filename: File to run the checksums against
    @type filename: String
    @param calc_prelink: Whether or not to reverse prelink before running the checksum
    @type calc_prelink: Integer
    @param strict: Enable/Disable strict checking (which stops exactly at a checksum failure and throws an exception)
    @type strict: Integer
    @rtype: Tuple
    @return: Result of the checks and possible message:
            1) If size fails, False, and a tuple containing a message, the given size, and the actual size
            2) If there is an os error, False, and a tuple containing the system error followed by 2 nulls
            3) If a checksum fails, False and a tuple containing a message, the given hash, and the actual hash
            4) If all checks succeed, return True and a fake reason
    """
    # Dict relates to single file only.
    # returns: (passed,reason)
    file_is_ok = True
    reason = "Reason unknown"
    try:
        mysize = os.stat(filename)[stat.ST_SIZE]
        if mydict.get("size") is not None and mydict["size"] != mysize:
            return False, (
                _("Filesize does not match recorded size"),
                mysize,
                mydict["size"],
            )
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise portage.exception.FileNotFound(filename)
        return False, (str(e), None, None)

    verifiable_hash_types = set(mydict).intersection(hashfunc_keys)
    verifiable_hash_types.discard("size")
    if not verifiable_hash_types:
        expected = set(hashfunc_keys)
        expected.discard("size")
        expected = list(expected)
        expected.sort()
        expected = " ".join(expected)
        got = set(mydict)
        got.discard("size")
        got = list(got)
        got.sort()
        got = " ".join(got)
        return False, (_("Insufficient data for checksum verification"), got, expected)

    for x in sorted(mydict):
        if x == "size":
            continue
        elif x in hashfunc_keys:
            myhash = perform_checksum(filename, x, calc_prelink=calc_prelink)[0]
            if mydict[x] != myhash:
                if strict:
                    raise portage.exception.DigestException(
                        f"Failed to verify '{filename}' on checksum type '{x}'"
                    )
                else:
                    file_is_ok = False
                    reason = (f"Failed on {x} verification", myhash, mydict[x])
                    break

    return file_is_ok, reason


def perform_checksum(filename, hashname="MD5", calc_prelink=0):
    """
    Run a specific checksum against a file. The filename can
    be either unicode or an encoded byte string. If filename
    is unicode then a UnicodeDecodeError will be raised if
    necessary.

    @param filename: File to run the checksum against
    @type filename: String
    @param hashname: The type of hash function to run
    @type hashname: String
    @param calc_prelink: Whether or not to reverse prelink before running the checksum
    @type calc_prelink: Integer
    @rtype: Tuple
    @return: The hash and size of the data
    """
    global prelink_capable
    # Make sure filename is encoded with the correct encoding before
    # it is passed to spawn (for prelink) and/or the hash function.
    filename = _unicode_encode(filename, encoding=_encodings["fs"], errors="strict")
    myfilename = filename
    prelink_tmpfile = None
    try:
        if calc_prelink and prelink_capable and is_prelinkable_elf(filename):
            # Create non-prelinked temporary file to checksum.
            # Files rejected by prelink are summed in place.
            try:
                tmpfile_fd, prelink_tmpfile = tempfile.mkstemp()
                try:
                    retval = portage.process.spawn(
                        [PRELINK_BINARY, "--verify", filename], fd_pipes={1: tmpfile_fd}
                    )
                finally:
                    os.close(tmpfile_fd)
                if retval == os.EX_OK:
                    myfilename = prelink_tmpfile
            except portage.exception.CommandNotFound:
                # This happens during uninstallation of prelink.
                prelink_capable = False
        try:
            if hashname not in hashfunc_keys:
                raise portage.exception.DigestException(
                    f"{hashname} hash function not available (needs dev-python/pycrypto)"
                )
            myhash, mysize = hashfunc_map[hashname].checksum_file(myfilename)
        except OSError as e:
            if e.errno in (errno.ENOENT, errno.ESTALE):
                raise portage.exception.FileNotFound(myfilename)
            elif e.errno == portage.exception.PermissionDenied.errno:
                raise portage.exception.PermissionDenied(myfilename)
            raise
        return myhash, mysize
    finally:
        if prelink_tmpfile:
            try:
                os.unlink(prelink_tmpfile)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise
                del e


def perform_multiple_checksums(filename, hashes=["MD5"], calc_prelink=0):
    """
    Run a group of checksums against a file.

    @param filename: File to run the checksums against
    @type filename: String
    @param hashes: A list of checksum functions to run against the file
    @type hashname: List
    @param calc_prelink: Whether or not to reverse prelink before running the checksum
    @type calc_prelink: Integer
    @rtype: Tuple
    @return: A dictionary in the form:
            return_value[hash_name] = (hash_result,size)
            for each given checksum
    """
    rVal = {}
    for x in hashes:
        if x not in hashfunc_keys:
            raise portage.exception.DigestException(
                f"{x} hash function not available (needs dev-python/pycrypto)"
            )
        rVal[x] = perform_checksum(filename, x, calc_prelink)[0]
    return rVal


def checksum_str(data, hashname="MD5"):
    """
    Run a specific checksum against a byte string.

    @param filename: Data to checksum
    @type filename: Bytes
    @param hashname: The type of hash function to run
    @type hashname: String
    @rtype: String
    @return: The hash (hex-digest) of the data
    """
    if hashname not in hashfunc_keys:
        raise portage.exception.DigestException(
            f"{hashname} hash function not available (needs dev-python/pycrypto)"
        )
    return hashfunc_map[hashname].checksum_str(data)



# Copyright 2018-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import re

import portage
from portage import os
from portage.const import GLOBAL_CONFIG_PATH

COMPAT_DISTDIR = "usr/portage/distfiles"
COMPAT_PKGDIR = "usr/portage/packages"
COMPAT_RPMDIR = "usr/portage/rpm"
COMPAT_MAIN_REPO = "usr/portage"


def main():
    """
    If the current installation is still configured to use any of the
    legacy default /usr/portage locations, then patch make.globals and
    repos.conf inside ${ED} to maintain compatible defaults. This is
    intended to be called from the ebuild as follows:

    pkg_preinst() {
            python_setup
            python_export PYTHON_SITEDIR
            env -u DISTDIR \
                    -u PORTAGE_OVERRIDE_EPREFIX \
                    -u PORTAGE_REPOSITORIES \
                    -u PORTDIR \
                    -u PORTDIR_OVERLAY \
                    PYTHONPATH="${ED%/}${PYTHON_SITEDIR}${PYTHONPATH:+:${PYTHONPATH}}" \
                    "${PYTHON}" -m portage._compat_upgrade.default_locations || die
    }
    """
    out = portage.output.EOutput()
    config = portage.settings

    compat_distdir = os.path.join(portage.const.EPREFIX or "/", COMPAT_DISTDIR)
    try:
        do_distdir = os.path.samefile(config["DISTDIR"], compat_distdir)
    except OSError:
        do_distdir = False

    compat_pkgdir = os.path.join(portage.const.EPREFIX or "/", COMPAT_PKGDIR)
    try:
        do_pkgdir = os.path.samefile(config["PKGDIR"], compat_pkgdir)
    except OSError:
        do_pkgdir = False

    compat_rpmdir = os.path.join(portage.const.EPREFIX or "/", COMPAT_RPMDIR)
    try:
        do_rpmdir = os.path.samefile(config["RPMDIR"], compat_rpmdir)
    except OSError:
        do_rpmdir = False

    compat_main_repo = os.path.join(portage.const.EPREFIX or "/", COMPAT_MAIN_REPO)
    try:
        do_main_repo = os.path.samefile(
            config.repositories.mainRepoLocation(), compat_main_repo
        )
    except OSError:
        do_main_repo = False

    if do_distdir or do_pkgdir or do_rpmdir:
        config_path = os.path.join(
            os.environ["ED"], GLOBAL_CONFIG_PATH.lstrip(os.sep), "make.globals"
        )
        with open(config_path) as f:
            content = f.read()
            if do_distdir:
                compat_setting = 'DISTDIR="{}"'.format(compat_distdir)
                out.einfo(
                    "Setting make.globals default {} for backward compatibility".format(
                        compat_setting
                    )
                )
                content = re.sub(
                    "^DISTDIR=.*$", compat_setting, content, flags=re.MULTILINE
                )
            if do_pkgdir:
                compat_setting = 'PKGDIR="{}"'.format(compat_pkgdir)
                out.einfo(
                    "Setting make.globals default {} for backward compatibility".format(
                        compat_setting
                    )
                )
                content = re.sub(
                    "^PKGDIR=.*$", compat_setting, content, flags=re.MULTILINE
                )
            if do_rpmdir:
                compat_setting = 'RPMDIR="{}"'.format(compat_rpmdir)
                out.einfo(
                    "Setting make.globals default {} for backward compatibility".format(
                        compat_setting
                    )
                )
                content = re.sub(
                    "^RPMDIR=.*$", compat_setting, content, flags=re.MULTILINE
                )
        with open(config_path, "w") as f:
            f.write(content)

    if do_main_repo:
        config_path = os.path.join(
            os.environ["ED"], GLOBAL_CONFIG_PATH.lstrip(os.sep), "repos.conf"
        )
        with open(config_path) as f:
            content = f.read()
            compat_setting = "location = {}".format(compat_main_repo)
            out.einfo(
                "Setting repos.conf default {} for backward compatibility".format(
                    compat_setting
                )
            )
            content = re.sub(
                "^location =.*$", compat_setting, content, flags=re.MULTILINE
            )
        with open(config_path, "w") as f:
            f.write(content)


if __name__ == "__main__":
    main()



# Copyright 2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from collections import OrderedDict
from collections.abc import Mapping
from hashlib import md5

from portage.localization import _
from portage.util import _recursive_file_list, writemsg
from portage.util.configparser import SafeConfigParser, ConfigParserError, read_configs


class BinRepoConfig:
    __slots__ = (
        "name",
        "name_fallback",
        "fetchcommand",
        "priority",
        "resumecommand",
        "sync_uri",
    )

    def __init__(self, opts):
        """
        Create a BinRepoConfig with options in opts.
        """
        for k in self.__slots__:
            setattr(self, k, opts.get(k.replace("_", "-")))

    def info_string(self):
        """
        Returns a formatted string containing informations about the repository.
        Used by emerge --info.
        """
        indent = " " * 4
        repo_msg = []
        repo_msg.append(self.name or self.name_fallback)
        if self.priority is not None:
            repo_msg.append(indent + "priority: " + str(self.priority))
        repo_msg.append(indent + "sync-uri: " + self.sync_uri)
        repo_msg.append("")
        return "\n".join(repo_msg)


class BinRepoConfigLoader(Mapping):
    def __init__(self, paths, settings):
        """Load config from files in paths"""

        # Defaults for value interpolation.
        parser_defaults = {
            "EPREFIX": settings["EPREFIX"],
            "EROOT": settings["EROOT"],
            "PORTAGE_CONFIGROOT": settings["PORTAGE_CONFIGROOT"],
            "ROOT": settings["ROOT"],
        }

        try:
            parser = self._parse(paths, parser_defaults)
        except ConfigParserError as e:
            writemsg(
                _("!!! Error while reading binrepo config file: %s\n") % e,
                noiselevel=-1,
            )
            parser = SafeConfigParser(defaults=parser_defaults)

        repos = []
        sync_uris = []
        for section_name in parser.sections():
            repo_data = dict(parser[section_name].items())
            repo_data["name"] = section_name
            repo = BinRepoConfig(repo_data)
            if repo.sync_uri is None:
                writemsg(
                    _("!!! Missing sync-uri setting for binrepo %s\n") % (repo.name,),
                    noiselevel=-1,
                )
                continue

            sync_uri = self._normalize_uri(repo.sync_uri)
            sync_uris.append(sync_uri)
            repo.sync_uri = sync_uri
            if repo.priority is not None:
                try:
                    repo.priority = int(repo.priority)
                except ValueError:
                    repo.priority = None
            repos.append(repo)

        sync_uris = set(sync_uris)
        current_priority = 0
        for sync_uri in reversed(settings.get("PORTAGE_BINHOST", "").split()):
            sync_uri = self._normalize_uri(sync_uri)
            if sync_uri not in sync_uris:
                current_priority += 1
                sync_uris.add(sync_uri)
                repos.append(
                    BinRepoConfig(
                        {
                            "name-fallback": self._digest_uri(sync_uri),
                            "name": None,
                            "priority": current_priority,
                            "sync-uri": sync_uri,
                        }
                    )
                )

        self._data = OrderedDict(
            (repo.name or repo.name_fallback, repo)
            for repo in sorted(
                repos,
                key=lambda repo: (repo.priority or 0, repo.name or repo.name_fallback),
            )
        )

    @staticmethod
    def _digest_uri(uri):
        return md5(uri.encode("utf_8")).hexdigest()

    @staticmethod
    def _normalize_uri(uri):
        return uri.rstrip("/")

    @staticmethod
    def _parse(paths, defaults):
        parser = SafeConfigParser(defaults=defaults)
        recursive_paths = []
        for p in paths:
            if isinstance(p, str):
                recursive_paths.extend(_recursive_file_list(p))
            else:
                recursive_paths.append(p)

        read_configs(parser, recursive_paths)
        return parser

    def __iter__(self):
        return iter(self._data)

    def __contains__(self, key):
        return key in self._data

    def __getitem__(self, key):
        return self._data[key]

    def __len__(self):
        return len(self._data)



# Copyright 2013-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import argparse
import logging
import sys

import portage
from portage import os
from portage.package.ebuild.fetch import ContentHashLayout
from portage.util import normalize_path, _recursive_file_list
from portage.util._async.run_main_scheduler import run_main_scheduler
from portage.util._async.SchedulerInterface import SchedulerInterface
from portage.util._eventloop.global_event_loop import global_event_loop
from .Config import Config
from .MirrorDistTask import MirrorDistTask


seconds_per_day = 24 * 60 * 60

common_options = (
    {
        "longopt": "--dry-run",
        "help": "perform a trial run with no changes made (usually combined "
        "with --verbose)",
        "action": "store_true",
    },
    {
        "longopt": "--verbose",
        "shortopt": "-v",
        "help": "display extra information on stderr "
        "(multiple occurences increase verbosity)",
        "action": "count",
        "default": 0,
    },
    {
        "longopt": "--ignore-default-opts",
        "help": "do not use the EMIRRORDIST_DEFAULT_OPTS environment variable",
        "action": "store_true",
    },
    {
        "longopt": "--distfiles",
        "help": "distfiles directory to use (required)",
        "metavar": "DIR",
    },
    {
        "longopt": "--jobs",
        "shortopt": "-j",
        "help": "number of concurrent jobs to run",
        "type": int,
    },
    {
        "longopt": "--load-average",
        "shortopt": "-l",
        "help": "load average limit for spawning of new concurrent jobs",
        "metavar": "LOAD",
        "type": float,
    },
    {
        "longopt": "--tries",
        "help": "maximum number of tries per file, 0 means unlimited (default is 10)",
        "default": 10,
        "type": int,
    },
    {"longopt": "--repo", "help": "name of repo to operate on"},
    {
        "longopt": "--config-root",
        "help": "location of portage config files",
        "metavar": "DIR",
    },
    {
        "longopt": "--repositories-configuration",
        "help": "override configuration of repositories (in format of repos.conf)",
    },
    {
        "longopt": "--strict-manifests",
        "help": 'manually override "strict" FEATURES setting',
        "choices": ("y", "n"),
        "metavar": "<y|n>",
    },
    {
        "longopt": "--failure-log",
        "help": "log file for fetch failures, with tab-delimited "
        "output, for reporting purposes",
        "metavar": "FILE",
    },
    {
        "longopt": "--success-log",
        "help": "log file for fetch successes, with tab-delimited "
        "output, for reporting purposes",
        "metavar": "FILE",
    },
    {
        "longopt": "--scheduled-deletion-log",
        "help": "log file for scheduled deletions, with tab-delimited "
        "output, for reporting purposes",
        "metavar": "FILE",
    },
    {
        "longopt": "--delete",
        "help": "enable deletion of unused distfiles",
        "action": "store_true",
    },
    {
        "longopt": "--deletion-db",
        "help": "database file used to track lifetime of files "
        "scheduled for delayed deletion",
        "metavar": "FILE",
    },
    {
        "longopt": "--deletion-delay",
        "help": "delay time for deletion, measured in seconds",
        "metavar": "SECONDS",
    },
    {
        "longopt": "--temp-dir",
        "help": "temporary directory for downloads",
        "metavar": "DIR",
    },
    {
        "longopt": "--mirror-overrides",
        "help": "file holding a list of mirror overrides",
        "metavar": "FILE",
    },
    {
        "longopt": "--mirror-skip",
        "help": "comma delimited list of mirror targets to skip " "when fetching",
    },
    {
        "longopt": "--restrict-mirror-exemptions",
        "help": "comma delimited list of mirror targets for which to "
        'ignore RESTRICT="mirror"',
    },
    {
        "longopt": "--verify-existing-digest",
        "help": "use digest as a verification of whether existing "
        "distfiles are valid",
        "action": "store_true",
    },
    {
        "longopt": "--distfiles-local",
        "help": "distfiles-local directory to use",
        "metavar": "DIR",
    },
    {
        "longopt": "--distfiles-db",
        "help": "database file used to track which ebuilds a " "distfile belongs to",
        "metavar": "FILE",
    },
    {
        "longopt": "--content-db",
        "help": "database file used to map content digests to"
        "distfiles names (required for content-hash layout)",
        "metavar": "FILE",
    },
    {
        "longopt": "--recycle-dir",
        "help": "directory for extended retention of files that "
        "are removed from distdir with the --delete option",
        "metavar": "DIR",
    },
    {
        "longopt": "--recycle-db",
        "help": "database file used to track lifetime of files " "in recycle dir",
        "metavar": "FILE",
    },
    {
        "longopt": "--recycle-deletion-delay",
        "help": "delay time for deletion of unused files from "
        "recycle dir, measured in seconds (defaults to "
        "the equivalent of 60 days)",
        "default": 60 * seconds_per_day,
        "metavar": "SECONDS",
        "type": int,
    },
    {
        "longopt": "--fetch-log-dir",
        "help": "directory for individual fetch logs",
        "metavar": "DIR",
    },
    {
        "longopt": "--whitelist-from",
        "help": "specifies a file containing a list of files to "
        "whitelist, one per line, # prefixed lines ignored",
        "action": "append",
        "metavar": "FILE",
    },
    {
        "longopt": "--symlinks",
        "help": "use symlinks rather than hardlinks for linking "
        "distfiles between layouts",
        "action": "store_true",
    },
    {
        "longopt": "--layout-conf",
        "help": "specifies layout.conf file to use instead of "
        "the one present in the distfiles directory",
        "metavar": "FILE",
    },
)


def parse_args(args):
    description = "emirrordist - a fetch tool for mirroring " "of package distfiles"
    usage = "emirrordist [options] <action>"
    parser = argparse.ArgumentParser(description=description, usage=usage)

    actions = parser.add_argument_group("Actions")
    actions.add_argument(
        "--version", action="store_true", help="display portage version and exit"
    )
    actions.add_argument(
        "--mirror",
        action="store_true",
        help="mirror distfiles for the selected repository",
    )

    common = parser.add_argument_group("Common options")
    for opt_info in common_options:
        opt_pargs = [opt_info["longopt"]]
        if opt_info.get("shortopt"):
            opt_pargs.append(opt_info["shortopt"])
        opt_kwargs = {"help": opt_info["help"]}
        for k in ("action", "choices", "default", "metavar", "type"):
            if k in opt_info:
                opt_kwargs[k] = opt_info[k]
        common.add_argument(*opt_pargs, **opt_kwargs)

    options, args = parser.parse_known_args(args)

    return (parser, options, args)


def emirrordist_main(args):

    # The calling environment is ignored, so the program is
    # completely controlled by commandline arguments.
    env = {}

    if not sys.stdout.isatty():
        portage.output.nocolor()
        env["NOCOLOR"] = "true"

    parser, options, args = parse_args(args)

    if options.version:
        sys.stdout.write("Portage %s\n" % portage.VERSION)
        return os.EX_OK

    config_root = options.config_root

    if options.repositories_configuration is not None:
        env["PORTAGE_REPOSITORIES"] = options.repositories_configuration

    settings = portage.config(config_root=config_root, local_config=False, env=env)

    default_opts = None
    if not options.ignore_default_opts:
        default_opts = settings.get("EMIRRORDIST_DEFAULT_OPTS", "").split()

    if default_opts:
        parser, options, args = parse_args(default_opts + args)

        settings = portage.config(config_root=config_root, local_config=False, env=env)

    if options.repo is None:
        if len(settings.repositories.prepos) == 2:
            for repo in settings.repositories:
                if repo.name != "DEFAULT":
                    options.repo = repo.name
                    break

        if options.repo is None:
            parser.error("--repo option is required")

    repo_path = settings.repositories.treemap.get(options.repo)
    if repo_path is None:
        parser.error("Unable to locate repository named '{}'".format(options.repo))

    if options.jobs is not None:
        options.jobs = int(options.jobs)

    if options.load_average is not None:
        options.load_average = float(options.load_average)

    if options.failure_log is not None:
        options.failure_log = normalize_path(os.path.abspath(options.failure_log))

        parent_dir = os.path.dirname(options.failure_log)
        if not (os.path.isdir(parent_dir) and os.access(parent_dir, os.W_OK | os.X_OK)):
            parser.error(
                ("--failure-log '%s' parent is not a " "writable directory")
                % options.failure_log
            )

    if options.success_log is not None:
        options.success_log = normalize_path(os.path.abspath(options.success_log))

        parent_dir = os.path.dirname(options.success_log)
        if not (os.path.isdir(parent_dir) and os.access(parent_dir, os.W_OK | os.X_OK)):
            parser.error(
                ("--success-log '%s' parent is not a " "writable directory")
                % options.success_log
            )

    if options.scheduled_deletion_log is not None:
        options.scheduled_deletion_log = normalize_path(
            os.path.abspath(options.scheduled_deletion_log)
        )

        parent_dir = os.path.dirname(options.scheduled_deletion_log)
        if not (os.path.isdir(parent_dir) and os.access(parent_dir, os.W_OK | os.X_OK)):
            parser.error(
                ("--scheduled-deletion-log '%s' parent is not a " "writable directory")
                % options.scheduled_deletion_log
            )

        if options.deletion_db is None:
            parser.error("--scheduled-deletion-log requires --deletion-db")

    if options.deletion_delay is not None:
        options.deletion_delay = int(options.deletion_delay)
        if options.deletion_db is None:
            parser.error("--deletion-delay requires --deletion-db")

    if options.deletion_db is not None:
        if options.deletion_delay is None:
            parser.error("--deletion-db requires --deletion-delay")
        options.deletion_db = normalize_path(os.path.abspath(options.deletion_db))

    if options.temp_dir is not None:
        options.temp_dir = normalize_path(os.path.abspath(options.temp_dir))

        if not (
            os.path.isdir(options.temp_dir)
            and os.access(options.temp_dir, os.W_OK | os.X_OK)
        ):
            parser.error(
                ("--temp-dir '%s' is not a " "writable directory") % options.temp_dir
            )

    if options.distfiles is not None:
        options.distfiles = normalize_path(os.path.abspath(options.distfiles))

        if not (
            os.path.isdir(options.distfiles)
            and os.access(options.distfiles, os.W_OK | os.X_OK)
        ):
            parser.error(
                ("--distfiles '%s' is not a " "writable directory") % options.distfiles
            )
    else:
        parser.error("missing required --distfiles parameter")

    if options.mirror_overrides is not None:
        options.mirror_overrides = normalize_path(
            os.path.abspath(options.mirror_overrides)
        )

        if not (
            os.access(options.mirror_overrides, os.R_OK)
            and os.path.isfile(options.mirror_overrides)
        ):
            parser.error(
                "--mirror-overrides-file '%s' is not a readable file"
                % options.mirror_overrides
            )

    if options.distfiles_local is not None:
        options.distfiles_local = normalize_path(
            os.path.abspath(options.distfiles_local)
        )

        if not (
            os.path.isdir(options.distfiles_local)
            and os.access(options.distfiles_local, os.W_OK | os.X_OK)
        ):
            parser.error(
                ("--distfiles-local '%s' is not a " "writable directory")
                % options.distfiles_local
            )

    if options.distfiles_db is not None:
        options.distfiles_db = normalize_path(os.path.abspath(options.distfiles_db))

    if options.tries is not None:
        options.tries = int(options.tries)

    if options.recycle_dir is not None:
        options.recycle_dir = normalize_path(os.path.abspath(options.recycle_dir))
        if not (
            os.path.isdir(options.recycle_dir)
            and os.access(options.recycle_dir, os.W_OK | os.X_OK)
        ):
            parser.error(
                ("--recycle-dir '%s' is not a " "writable directory")
                % options.recycle_dir
            )

    if options.recycle_db is not None:
        if options.recycle_dir is None:
            parser.error("--recycle-db requires " "--recycle-dir to be specified")
        options.recycle_db = normalize_path(os.path.abspath(options.recycle_db))

    if options.recycle_deletion_delay is not None:
        options.recycle_deletion_delay = int(options.recycle_deletion_delay)

    if options.fetch_log_dir is not None:
        options.fetch_log_dir = normalize_path(os.path.abspath(options.fetch_log_dir))

        if not (
            os.path.isdir(options.fetch_log_dir)
            and os.access(options.fetch_log_dir, os.W_OK | os.X_OK)
        ):
            parser.error(
                ("--fetch-log-dir '%s' is not a " "writable directory")
                % options.fetch_log_dir
            )

    if options.whitelist_from:
        normalized_paths = []
        for x in options.whitelist_from:
            path = normalize_path(os.path.abspath(x))
            if not os.access(path, os.R_OK):
                parser.error("--whitelist-from '%s' is not readable" % x)
            if os.path.isfile(path):
                normalized_paths.append(path)
            elif os.path.isdir(path):
                for file in _recursive_file_list(path):
                    if not os.access(file, os.R_OK):
                        parser.error(
                            "--whitelist-from '%s' directory contains not readable file '%s'"
                            % (x, file)
                        )
                    normalized_paths.append(file)
            else:
                parser.error(
                    "--whitelist-from '%s' is not a regular file or a directory" % x
                )
        options.whitelist_from = normalized_paths

    if options.strict_manifests is not None:
        if options.strict_manifests == "y":
            settings.features.add("strict")
        else:
            settings.features.discard("strict")

    settings.lock()

    portdb = portage.portdbapi(mysettings=settings)

    # Limit ebuilds to the specified repo.
    portdb.porttrees = [repo_path]

    portage.util.initialize_logger()

    if options.verbose > 0:
        l = logging.getLogger()
        l.setLevel(l.getEffectiveLevel() - 10 * options.verbose)

    with Config(options, portdb, SchedulerInterface(global_event_loop())) as config:

        if not options.mirror:
            parser.error("No action specified")

        if options.delete and config.content_db is None:
            for layout in config.layouts:
                if isinstance(layout, ContentHashLayout):
                    parser.error(
                        "content-hash layout requires " "--content-db to be specified"
                    )

        returncode = os.EX_OK

        if options.mirror:
            signum = run_main_scheduler(MirrorDistTask(config))
            if signum is not None:
                sys.exit(128 + signum)

    return returncode



#!/usr/bin/env python
# Copyright 2011-2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

doc = """Helper tool for converting installed files to custom prefixes.

In other words, eprefixy ${D} for Gentoo/Prefix."""
__doc__ = doc

import argparse
import os
import stat
import sys

CONTENT_ENCODING = "utf_8"
FS_ENCODING = "utf_8"

try:
    import magic
except ImportError:
    magic = None
else:
    try:
        magic.MIME_TYPE
    except AttributeError:
        # magic module seems to be broken
        magic = None


class IsTextFile:
    def __init__(self):
        if magic is not None:
            self._call = self._is_text_magic
            self._m = magic.open(magic.MIME_TYPE)
            self._m.load()
        else:
            self._call = self._is_text_encoding
            self._encoding = CONTENT_ENCODING

    def __call__(self, filename):
        """
        Returns True if the given file is a text file, and False otherwise.
        """
        return self._call(filename)

    def _is_text_magic(self, filename):
        mime_type = self._m.file(filename)
        if isinstance(mime_type, bytes):
            mime_type = mime_type.decode("ascii", "replace")
        return mime_type.startswith("text/")

    def _is_text_encoding(self, filename):
        try:
            for line in open(filename, encoding=self._encoding):
                pass
        except UnicodeDecodeError:
            return False
        return True


def chpath_inplace(filename, is_text_file, old, new):
    """
    Returns True if any modifications were made, and False otherwise.
    """

    modified = False
    orig_stat = os.lstat(filename)
    try:
        f = open(filename, buffering=0, mode="r+b")
    except OSError:
        try:
            orig_mode = stat.S_IMODE(os.lstat(filename).st_mode)
        except OSError as e:
            sys.stderr.write("{}: {}\n".format(e, filename))
            return
        temp_mode = 0o200 | orig_mode
        os.chmod(filename, temp_mode)
        try:
            f = open(filename, buffering=0, mode="r+b")
        finally:
            os.chmod(filename, orig_mode)

    len_old = len(old)
    len_new = len(new)
    matched_byte_count = 0
    while True:
        in_byte = f.read(1)

        if not in_byte:
            break

        if in_byte == old[matched_byte_count : matched_byte_count + 1]:
            matched_byte_count += 1
            if matched_byte_count == len_old:
                modified = True
                matched_byte_count = 0
                end_position = f.tell()
                start_position = end_position - len_old
                if not is_text_file:
                    # search backwards for leading slashes written by
                    # a previous invocation of this tool
                    num_to_write = len_old
                    f.seek(start_position - 1)
                    while True:
                        if f.read(1) != b"/":
                            break
                        num_to_write += 1
                        f.seek(f.tell() - 2)

                    # pad with as many leading slashes as necessary
                    while num_to_write > len_new:
                        f.write(b"/")
                        num_to_write -= 1
                    f.write(new)
                else:
                    remainder = f.read()
                    f.seek(start_position)
                    f.write(new)
                    if remainder:
                        f.write(remainder)
                        f.truncate()
                        f.seek(start_position + len_new)
        elif matched_byte_count > 0:
            # back up an try to start a new match after
            # the first byte of the previous partial match
            f.seek(f.tell() - matched_byte_count)
            matched_byte_count = 0

    f.close()
    if modified:
        orig_mtime = orig_stat.st_mtime_ns
        os.utime(filename, ns=(orig_mtime, orig_mtime))
    return modified


def chpath_inplace_symlink(filename, st, old, new):
    target = os.readlink(filename)
    if target.startswith(old):
        new_target = new + target[len(old) :]
        os.unlink(filename)
        os.symlink(new_target, filename)
        os.lchown(filename, st.st_uid, st.st_gid)


def main(argv):

    parser = argparse.ArgumentParser(description=doc)
    parser.add_argument("location", default=None, help="root directory (e.g. ${D})")
    parser.add_argument("old", default=None, help="original build prefix (e.g. /)")
    parser.add_argument(
        "new", default=None, help="new install prefix (e.g. ${EPREFIX})"
    )
    opts = parser.parse_args(argv)

    location, old, new = opts.location, opts.old, opts.new

    is_text_file = IsTextFile()

    if not isinstance(location, bytes):
        location = location.encode(FS_ENCODING)
    if not isinstance(old, bytes):
        old = old.encode(FS_ENCODING)
    if not isinstance(new, bytes):
        new = new.encode(FS_ENCODING)

    st = os.lstat(location)

    if stat.S_ISDIR(st.st_mode):
        for parent, dirs, files in os.walk(location):
            for filename in files:
                filename = os.path.join(parent, filename)
                try:
                    st = os.lstat(filename)
                except OSError:
                    pass
                else:
                    if stat.S_ISREG(st.st_mode):
                        chpath_inplace(filename, is_text_file(filename), old, new)
                    elif stat.S_ISLNK(st.st_mode):
                        chpath_inplace_symlink(filename, st, old, new)

    elif stat.S_ISREG(st.st_mode):
        chpath_inplace(location, is_text_file(location), old, new)
    elif stat.S_ISLNK(st.st_mode):
        chpath_inplace_symlink(location, st, old, new)

    return os.EX_OK


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/env python
# Copyright 2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import argparse
import sys
import portage

portage._internal_caller = True
from portage import os
from portage.output import EOutput


def command_compose(args):
    eout = EOutput()

    usage = "usage: compose <package_cpv> <binpkg_path> <metadata_dir> <image_dir>\n"

    if len(args) != 4:
        sys.stderr.write(usage)
        sys.stderr.write("4 arguments are required, got %s\n" % len(args))
        return 1

    basename, binpkg_path, metadata_dir, image_dir = args

    if not os.path.isdir(metadata_dir):
        sys.stderr.write(usage)
        sys.stderr.write("Argument 3 is not a directory: '%s'\n" % metadata_dir)
        return 1

    if not os.path.isdir(image_dir):
        sys.stderr.write(usage)
        sys.stderr.write("Argument 4 is not a directory: '%s'\n" % image_dir)
        return 1

    try:
        gpkg_file = portage.gpkg.gpkg(portage.settings, basename, binpkg_path)
        metadata = gpkg_file._generate_metadata_from_dir(metadata_dir)
        gpkg_file.compress(image_dir, metadata)
    except portage.exception.CompressorOperationFailed:
        eout.eerror("Compressor Operation Failed")
        exit(1)
    return os.EX_OK


def main(argv):

    if argv and isinstance(argv[0], bytes):
        for i, x in enumerate(argv):
            argv[i] = portage._unicode_decode(x, errors="strict")

    valid_commands = ("compress",)
    description = "Perform metadata operations on a binary package."
    usage = "usage: %s COMMAND [args]" % os.path.basename(argv[0])

    parser = argparse.ArgumentParser(description=description, usage=usage)
    options, args = parser.parse_known_args(argv[1:])

    if not args:
        parser.error("missing command argument")

    command = args[0]

    if command not in valid_commands:
        parser.error("invalid command: '%s'" % command)

    if command == "compress":
        rval = command_compose(args[1:])
    else:
        raise AssertionError("invalid command: '%s'" % command)

    return rval


if __name__ == "__main__":
    rval = main(sys.argv[:])
    sys.exit(rval)



# Copyright 1998-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import json
import logging
import pickle
import stat

from portage import abssymlink
from portage import os
from portage import _encodings
from portage import _os_merge
from portage import _unicode_decode
from portage import _unicode_encode
from portage.exception import PermissionDenied
from portage.localization import _
from portage.util import atomic_ofstream
from portage.util import writemsg_level
from portage.versions import cpv_getkey
from portage.locks import lockfile, unlockfile


class PreservedLibsRegistry:
    """This class handles the tracking of preserved library objects"""

    # JSON read support has been available since portage-2.2.0_alpha89.
    _json_write = True

    _json_write_opts = {
        "ensure_ascii": False,
        "indent": "\t",
        "sort_keys": True,
    }

    def __init__(self, root, filename):
        """
            @param root: root used to check existence of paths in pruneNonExisting
        @type root: String
            @param filename: absolute path for saving the preserved libs records
        @type filename: String
        """
        self._root = root
        self._filename = filename
        self._data = None
        self._lock = None

    def lock(self):
        """Grab an exclusive lock on the preserved libs registry."""
        if self._lock is not None:
            raise AssertionError("already locked")
        self._lock = lockfile(self._filename)

    def unlock(self):
        """Release our exclusive lock on the preserved libs registry."""
        if self._lock is None:
            raise AssertionError("not locked")
        unlockfile(self._lock)
        self._lock = None

    def load(self):
        """Reload the registry data from file"""
        self._data = None
        f = None
        content = None
        try:
            f = open(
                _unicode_encode(
                    self._filename, encoding=_encodings["fs"], errors="strict"
                ),
                "rb",
            )
            content = f.read()
        except OSError as e:
            if not hasattr(e, "errno"):
                raise
            elif e.errno == errno.ENOENT:
                pass
            elif e.errno == PermissionDenied.errno:
                raise PermissionDenied(self._filename)
            else:
                raise
        finally:
            if f is not None:
                f.close()

        # content is empty if it's an empty lock file
        if content:
            try:
                self._data = json.loads(
                    _unicode_decode(
                        content, encoding=_encodings["repo.content"], errors="strict"
                    )
                )
            except SystemExit:
                raise
            except Exception as e:
                try:
                    self._data = pickle.loads(content)
                except SystemExit:
                    raise
                except Exception:
                    writemsg_level(
                        _("!!! Error loading '%s': %s\n") % (self._filename, e),
                        level=logging.ERROR,
                        noiselevel=-1,
                    )

        if self._data is None:
            self._data = {}
        else:
            for k, v in self._data.items():
                if (
                    isinstance(v, (list, tuple))
                    and len(v) == 3
                    and isinstance(v[2], set)
                ):
                    # convert set to list, for write with JSONEncoder
                    self._data[k] = (v[0], v[1], list(v[2]))

        self._data_orig = self._data.copy()
        self.pruneNonExisting()

    def store(self):
        """
        Store the registry data to the file. The existing inode will be
        replaced atomically, so if that inode is currently being used
        for a lock then that lock will be rendered useless. Therefore,
        it is important not to call this method until the current lock
        is ready to be immediately released.
        """
        if os.environ.get("SANDBOX_ON") == "1" or self._data == self._data_orig:
            return
        try:
            f = atomic_ofstream(self._filename, "wb")
            if self._json_write:
                f.write(
                    _unicode_encode(
                        json.dumps(self._data, **self._json_write_opts),
                        encoding=_encodings["repo.content"],
                        errors="strict",
                    )
                )
            else:
                pickle.dump(self._data, f, protocol=2)
            f.close()
        except OSError as e:
            if e.errno != PermissionDenied.errno:
                writemsg_level(
                    "!!! {} {}\n".format(e, self._filename),
                    level=logging.ERROR,
                    noiselevel=-1,
                )
        else:
            self._data_orig = self._data.copy()

    def _normalize_counter(self, counter):
        """
        For simplicity, normalize as a unicode string
        and strip whitespace. This avoids the need for
        int conversion and a possible ValueError resulting
        from vardb corruption.
        """
        if not isinstance(counter, str):
            counter = str(counter)
        return _unicode_decode(counter).strip()

    def register(self, cpv, slot, counter, paths):
        """Register new objects in the registry. If there is a record with the
        same packagename (internally derived from cpv) and slot it is
        overwritten with the new data.
        @param cpv: package instance that owns the objects
        @type cpv: CPV (as String)
        @param slot: the value of SLOT of the given package instance
        @type slot: String
        @param counter: vdb counter value for the package instance
        @type counter: String
        @param paths: absolute paths of objects that got preserved during an update
        @type paths: List
        """
        cp = cpv_getkey(cpv)
        cps = cp + ":" + slot
        counter = self._normalize_counter(counter)
        if (
            len(paths) == 0
            and cps in self._data
            and self._data[cps][0] == cpv
            and self._normalize_counter(self._data[cps][1]) == counter
        ):
            del self._data[cps]
        elif len(paths) > 0:
            if isinstance(paths, set):
                # convert set to list, for write with JSONEncoder
                paths = list(paths)
            self._data[cps] = (cpv, counter, paths)

    def unregister(self, cpv, slot, counter):
        """Remove a previous registration of preserved objects for the given package.
        @param cpv: package instance whose records should be removed
        @type cpv: CPV (as String)
        @param slot: the value of SLOT of the given package instance
        @type slot: String
        """
        self.register(cpv, slot, counter, [])

    def pruneNonExisting(self):
        """Remove all records for objects that no longer exist on the filesystem."""

        os = _os_merge

        for cps in list(self._data):
            cpv, counter, _paths = self._data[cps]

            paths = []
            hardlinks = set()
            symlinks = {}
            for f in _paths:
                f_abs = os.path.join(self._root, f.lstrip(os.sep))
                try:
                    lst = os.lstat(f_abs)
                except OSError:
                    continue
                if stat.S_ISLNK(lst.st_mode):
                    try:
                        symlinks[f] = os.readlink(f_abs)
                    except OSError:
                        continue
                elif stat.S_ISREG(lst.st_mode):
                    hardlinks.add(f)
                    paths.append(f)

            # Only count symlinks as preserved if they still point to a hardink
            # in the same directory, in order to handle cases where a tool such
            # as eselect-opengl has updated the symlink to point to a hardlink
            # in a different directory (see bug #406837). The unused hardlink
            # is automatically found by _find_unused_preserved_libs, since the
            # soname symlink no longer points to it. After the hardlink is
            # removed by _remove_preserved_libs, it calls pruneNonExisting
            # which eliminates the irrelevant symlink from the registry here.
            for f, target in symlinks.items():
                if abssymlink(f, target=target) in hardlinks:
                    paths.append(f)

            if len(paths) > 0:
                self._data[cps] = (cpv, counter, paths)
            else:
                del self._data[cps]

    def hasEntries(self):
        """Check if this registry contains any records."""
        if self._data is None:
            self.load()
        return len(self._data) > 0

    def getPreservedLibs(self):
        """Return a mapping of packages->preserved objects.
        @return mapping of package instances to preserved objects
        @rtype Dict cpv->list-of-paths
        """
        if self._data is None:
            self.load()
        rValue = {}
        for cps in self._data:
            rValue[self._data[cps][0]] = self._data[cps][2]
        return rValue



# Copyright 2005-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

doc = """Check and clean the config tracker list for uninstalled packages."""
__doc__ = doc


module_spec = {
    "name": "config",
    "description": doc,
    "provides": {
        "module1": {
            "name": "cleanconfmem",
            "sourcefile": "config",
            "class": "CleanConfig",
            "description": doc,
            "functions": ["check", "fix"],
            "func_desc": {},
        }
    },
}



# Copyright 2009-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["lazyimport"]

import sys
import types

try:
    import threading
except ImportError:
    import dummy_threading as threading

from portage.proxy.objectproxy import ObjectProxy


_module_proxies = {}
_module_proxies_lock = threading.RLock()


def _preload_portage_submodules():
    """
    Load lazily referenced portage submodules into memory,
    so imports won't fail during portage upgrade/downgrade.
    Note that this recursively loads only the modules that
    are lazily referenced by currently imported modules,
    so some portage submodules may still remain unimported
    after this function is called.
    """
    imported = set()
    while True:
        remaining = False
        for name in list(_module_proxies):
            if name.startswith("portage.") or name.startswith("_emerge."):
                if name in imported:
                    continue
                imported.add(name)
                remaining = True
                __import__(name)
                _unregister_module_proxy(name)
        if not remaining:
            break


def _register_module_proxy(name, proxy):
    _module_proxies_lock.acquire()
    try:
        proxy_list = _module_proxies.get(name)
        if proxy_list is None:
            proxy_list = []
            _module_proxies[name] = proxy_list
        proxy_list.append(proxy)
    finally:
        _module_proxies_lock.release()


def _unregister_module_proxy(name):
    """
    Destroy all proxies that reference the give module name. Also, check
    for other proxies referenced by modules that have been imported and
    destroy those proxies too. This way, destruction of a single proxy
    can trigger destruction of all the rest. If a target module appears
    to be partially imported (indicated when an AttributeError is caught),
    this function will leave in place proxies that reference it.
    """
    _module_proxies_lock.acquire()
    try:
        if name in _module_proxies:
            modules = sys.modules
            for name, proxy_list in list(_module_proxies.items()):
                if name not in modules:
                    continue
                # First delete this name from the dict so that
                # if this same thread reenters below, it won't
                # enter this path again.
                del _module_proxies[name]
                try:
                    while proxy_list:
                        proxy = proxy_list.pop()
                        object.__getattribute__(proxy, "_get_target")()
                except AttributeError:
                    # Apparently the target module is only partially
                    # imported, so proxies that reference it cannot
                    # be destroyed yet.
                    proxy_list.append(proxy)
                    _module_proxies[name] = proxy_list
    finally:
        _module_proxies_lock.release()


class _LazyImport(ObjectProxy):

    __slots__ = ("_scope", "_alias", "_name", "_target")

    def __init__(self, scope, alias, name):
        ObjectProxy.__init__(self)
        object.__setattr__(self, "_scope", scope)
        object.__setattr__(self, "_alias", alias)
        object.__setattr__(self, "_name", name)
        _register_module_proxy(name, self)

    def _get_target(self):
        try:
            return object.__getattribute__(self, "_target")
        except AttributeError:
            pass
        name = object.__getattribute__(self, "_name")
        __import__(name)
        target = sys.modules[name]
        object.__setattr__(self, "_target", target)
        object.__getattribute__(self, "_scope")[
            object.__getattribute__(self, "_alias")
        ] = target
        _unregister_module_proxy(name)
        return target


class _LazyImportFrom(_LazyImport):

    __slots__ = ("_attr_name",)

    def __init__(self, scope, name, attr_name, alias):
        object.__setattr__(self, "_attr_name", attr_name)
        _LazyImport.__init__(self, scope, alias, name)

    def _get_target(self):
        try:
            return object.__getattribute__(self, "_target")
        except AttributeError:
            pass
        name = object.__getattribute__(self, "_name")
        attr_name = object.__getattribute__(self, "_attr_name")
        __import__(name)
        try:
            target = getattr(sys.modules[name], attr_name)
        except AttributeError:
            # Try to import it as a submodule
            try:
                __import__("{}.{}".format(name, attr_name))
            except ImportError:
                pass
            # If it's a submodule, this will succeed. Otherwise, it may
            # be that the module is only partially imported, so raise
            # AttributeError for _unregister_module_proxy() to handle.
            target = getattr(sys.modules[name], attr_name)

        object.__setattr__(self, "_target", target)
        object.__getattribute__(self, "_scope")[
            object.__getattribute__(self, "_alias")
        ] = target
        _unregister_module_proxy(name)
        return target


def lazyimport(scope, *args):
    """
    Create a proxy in the given scope in order to performa a lazy import.

    Syntax         Result
    foo            import foo
    foo:bar,baz    from foo import bar, baz
    foo:bar@baz    from foo import bar as baz

    @param scope: the scope in which to place the import, typically globals()
    @type myfilename: dict
    @param args: module names to import
    @type args: strings
    """

    modules = sys.modules

    for s in args:
        parts = s.split(":", 1)
        if len(parts) == 1:
            name = s

            if not name or not isinstance(name, str):
                raise ValueError(name)

            components = name.split(".")
            parent_scope = scope
            for i in range(len(components)):
                alias = components[i]
                if i < len(components) - 1:
                    parent_name = ".".join(components[: i + 1])
                    __import__(parent_name)
                    mod = modules.get(parent_name)
                    if not isinstance(mod, types.ModuleType):
                        # raise an exception
                        __import__(name)
                    parent_scope[alias] = mod
                    parent_scope = mod.__dict__
                    continue

                already_imported = modules.get(name)
                if already_imported is not None:
                    parent_scope[alias] = already_imported
                else:
                    parent_scope[alias] = _LazyImport(parent_scope, alias, name)

        else:
            name, fromlist = parts
            already_imported = modules.get(name)
            fromlist = fromlist.split(",")
            for s in fromlist:
                if not s:
                    # This happens if there's an extra comma in fromlist.
                    raise ValueError("Empty module attribute name")
                alias = s.split("@", 1)
                if len(alias) == 1:
                    alias = alias[0]
                    attr_name = alias
                else:
                    attr_name, alias = alias
                if already_imported is not None:
                    try:
                        scope[alias] = getattr(already_imported, attr_name)
                    except AttributeError:
                        # Apparently the target module is only partially
                        # imported, so create a proxy.
                        already_imported = None
                        scope[alias] = _LazyImportFrom(scope, name, attr_name, alias)
                else:
                    scope[alias] = _LazyImportFrom(scope, name, attr_name, alias)



# Copyright 2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import asyncio

from portage.tests import TestCase
from portage.util.futures.unix_events import DefaultEventLoopPolicy


class PolicyWrapperRecursionTestCase(TestCase):
    def testPolicyWrapperRecursion(self):
        initial_policy = asyncio.get_event_loop_policy()
        if not isinstance(initial_policy, DefaultEventLoopPolicy):
            asyncio.set_event_loop_policy(DefaultEventLoopPolicy())

        try:
            with self.assertRaises(NotImplementedError):
                asyncio.get_event_loop()

            with self.assertRaises(NotImplementedError):
                asyncio.get_child_watcher()
        finally:
            asyncio.set_event_loop_policy(initial_policy)



# Copyright 2012-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import io
from datetime import datetime
from time import mktime
from email.utils import formatdate, parsedate
from urllib.request import urlopen as _urlopen
import urllib.parse as urllib_parse
import urllib.request as urllib_request


# to account for the difference between TIMESTAMP of the index' contents
#  and the file-'mtime'
TIMESTAMP_TOLERANCE = 5


def have_pep_476():
    """
    Test whether ssl certificate verification is enabled by default for
    stdlib http clients (PEP 476).

    @returns: bool, True if ssl certificate verification is enabled by
            default
    """
    return hasattr(__import__("ssl"), "_create_unverified_context")


def urlopen(url, if_modified_since=None, headers={}, proxies=None):
    parse_result = urllib_parse.urlparse(url)
    if parse_result.scheme not in ("http", "https"):
        return _urlopen(url)

    netloc = parse_result.netloc.rpartition("@")[-1]
    url = urllib_parse.urlunparse(
        (
            parse_result.scheme,
            netloc,
            parse_result.path,
            parse_result.params,
            parse_result.query,
            parse_result.fragment,
        )
    )
    password_manager = urllib_request.HTTPPasswordMgrWithDefaultRealm()
    request = urllib_request.Request(url)
    request.add_header("User-Agent", "Gentoo Portage")
    for key in headers:
        request.add_header(key, headers[key])
    if if_modified_since:
        request.add_header("If-Modified-Since", _timestamp_to_http(if_modified_since))
    if parse_result.username is not None:
        password_manager.add_password(
            None, url, parse_result.username, parse_result.password
        )

    handlers = [CompressedResponseProcessor(password_manager)]
    if proxies:
        handlers.append(urllib_request.ProxyHandler(proxies))
    opener = urllib_request.build_opener(*handlers)

    hdl = opener.open(request)
    if hdl.headers.get("last-modified", ""):
        try:
            add_header = hdl.headers.add_header
        except AttributeError:
            # Python 2
            add_header = hdl.headers.addheader
        add_header("timestamp", _http_to_timestamp(hdl.headers.get("last-modified")))
    return hdl


def _timestamp_to_http(timestamp):
    dt = datetime.fromtimestamp(float(int(timestamp) + TIMESTAMP_TOLERANCE))
    stamp = mktime(dt.timetuple())
    return formatdate(timeval=stamp, localtime=False, usegmt=True)


def _http_to_timestamp(http_datetime_string):
    timestamp = mktime(parsedate(http_datetime_string))
    return str(int(timestamp))


class CompressedResponseProcessor(urllib_request.HTTPBasicAuthHandler):
    # Handler for compressed responses.

    def http_request(self, req):
        req.add_header("Accept-Encoding", "bzip2,gzip,deflate")
        return req

    https_request = http_request

    def http_response(self, req, response):
        decompressed = None
        if response.headers.get("content-encoding") == "bzip2":
            import bz2

            decompressed = io.BytesIO(bz2.decompress(response.read()))
        elif response.headers.get("content-encoding") == "gzip":
            from gzip import GzipFile

            decompressed = GzipFile(fileobj=io.BytesIO(response.read()), mode="r")
        elif response.headers.get("content-encoding") == "deflate":
            import zlib

            try:
                decompressed = io.BytesIO(zlib.decompress(response.read()))
            except zlib.error:  # they ignored RFC1950
                decompressed = io.BytesIO(
                    zlib.decompress(response.read(), -zlib.MAX_WBITS)
                )
        if decompressed:
            old_response = response
            response = urllib_request.addinfourl(
                decompressed, old_response.headers, old_response.url, old_response.code
            )
            response.msg = old_response.msg
        return response

    https_response = http_response



# Copyright 2010-2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import os as _os
import re

from portage import _unicode_decode
from portage.exception import InvalidData

#########################################################
# 	This an re-implementaion of dev-util/lafilefixer-0.5.
# 	rewrite_lafile() takes the contents of an lafile as a string
# 	It then parses the dependency_libs and inherited_linker_flags
# 	entries.
# 	We insist on dependency_libs being present. inherited_linker_flags
# 	is optional.
# 	There are strict rules about the syntax imposed by libtool's libltdl.
# 	See 'parse_dotla_file' and 'trim' functions in libltdl/ltdl.c.
# 	Note that duplicated entries of dependency_libs and inherited_linker_flags
# 	are ignored by libtool (last one wins), but we treat it as error (like
# 	lafilefixer does).
# 	What it does:
# 		* Replaces all .la files with absolut paths in dependency_libs with
# 		  corresponding -l* and -L* entries
# 		  (/usr/lib64/libfoo.la -> -L/usr/lib64 -lfoo)
# 		* Moves various flags (see flag_re below) to inherited_linker_flags,
# 		  if such an entry was present.
# 		* Reorders dependency_libs such that all -R* entries precede -L* entries
# 		  and these precede all other entries.
# 		* Remove duplicated entries from dependency_libs
# 		* Takes care that no entry to inherited_linker_flags is added that is
# 		  already there.
#########################################################

# These regexes are used to parse the interesting entries in the la file
dep_libs_re = re.compile(b"dependency_libs='(?P<value>[^']*)'$")
inh_link_flags_re = re.compile(b"inherited_linker_flags='(?P<value>[^']*)'$")

# regexes for replacing stuff in -L entries.
# replace 'X11R6/lib' and 'local/lib' with 'lib', no idea what's this about.
X11_local_sub = re.compile(b"X11R6/lib|local/lib")
# get rid of the '..'
pkgconfig_sub1 = re.compile(rb"usr/lib[^/]*/pkgconfig/\.\./\.\.")
pkgconfig_sub2 = re.compile(rb"(?P<usrlib>usr/lib[^/]*)/pkgconfig/\.\.")

# detect flags that should go into inherited_linker_flags instead of dependency_libs
flag_re = re.compile(
    b"-mt|-mthreads|-kthread|-Kthread|-pthread|-pthreads|--thread-safe|-threads"
)


def _parse_lafile_contents(contents):
    """
    Parses 'dependency_libs' and 'inherited_linker_flags' lines.
    """

    dep_libs = None
    inh_link_flags = None

    for line in contents.split(b"\n"):
        m = dep_libs_re.match(line)
        if m:
            if dep_libs is not None:
                raise InvalidData("duplicated dependency_libs entry")
            dep_libs = m.group("value")
            continue

        m = inh_link_flags_re.match(line)
        if m:
            if inh_link_flags is not None:
                raise InvalidData("duplicated inherited_linker_flags entry")
            inh_link_flags = m.group("value")
            continue

    return dep_libs, inh_link_flags


def rewrite_lafile(contents):
    """
    Given the contents of an .la file, parse and fix it.
    This operates with strings of raw bytes (assumed to contain some ascii
    characters), in order to avoid any potential character encoding issues.
    Raises 'InvalidData' if the .la file is invalid.
    @param contents: the contents of a libtool archive file
    @type contents: bytes
    @rtype: tuple
    @return: (True, fixed_contents) if something needed to be
            fixed, (False, None) otherwise.
    """
    # Parse the 'dependency_libs' and 'inherited_linker_flags' lines.
    dep_libs, inh_link_flags = _parse_lafile_contents(contents)

    if dep_libs is None:
        raise InvalidData("missing or invalid dependency_libs")

    new_dep_libs = []
    new_inh_link_flags = []
    librpath = []
    libladir = []

    if inh_link_flags is not None:
        new_inh_link_flags = inh_link_flags.split()

    # Check entries in 'dependency_libs'.
    for dep_libs_entry in dep_libs.split():
        if dep_libs_entry.startswith(b"-l"):
            # -lfoo, keep it
            if dep_libs_entry not in new_dep_libs:
                new_dep_libs.append(dep_libs_entry)

        elif dep_libs_entry.endswith(b".la"):
            # Two cases:
            # 1) /usr/lib64/libfoo.la, turn it into -lfoo and append -L/usr/lib64 to libladir
            # 2) libfoo.la, keep it
            dirname, basename = _os.path.split(dep_libs_entry)

            if not dirname or not basename.startswith(b"lib"):
                if dep_libs_entry not in new_dep_libs:
                    new_dep_libs.append(dep_libs_entry)
            else:
                # /usr/lib64/libfoo.la -> -lfoo
                lib = b"-l" + basename[3:-3]
                if lib not in new_dep_libs:
                    new_dep_libs.append(lib)
                # /usr/lib64/libfoo.la -> -L/usr/lib64
                ladir = b"-L" + dirname
                if ladir not in libladir:
                    libladir.append(ladir)

        elif dep_libs_entry.startswith(b"-L"):
            # Do some replacement magic and store them in 'libladir'.
            # This allows us to place all -L entries at the beginning
            # of 'dependency_libs'.
            ladir = dep_libs_entry

            ladir = X11_local_sub.sub(b"lib", ladir)
            ladir = pkgconfig_sub1.sub(b"usr", ladir)
            ladir = pkgconfig_sub2.sub(rb"\g<usrlib>", ladir)

            if ladir not in libladir:
                libladir.append(ladir)

        elif dep_libs_entry.startswith(b"-R"):
            if dep_libs_entry not in librpath:
                librpath.append(dep_libs_entry)

        elif flag_re.match(dep_libs_entry):
            # All this stuff goes into inh_link_flags, if the la file has such an entry.
            # If it doesn't, they stay in 'dependency_libs'.
            if inh_link_flags is not None:
                if dep_libs_entry not in new_inh_link_flags:
                    new_inh_link_flags.append(dep_libs_entry)
            else:
                if dep_libs_entry not in new_dep_libs:
                    new_dep_libs.append(dep_libs_entry)

        else:
            raise InvalidData(
                "Error: Unexpected entry '%s' in 'dependency_libs'"
                % _unicode_decode(dep_libs_entry)
            )

    # What should 'dependency_libs' and 'inherited_linker_flags' look like?
    expected_dep_libs = b""
    for x in (librpath, libladir, new_dep_libs):
        if x:
            expected_dep_libs += b" " + b" ".join(x)

    expected_inh_link_flags = b""
    if new_inh_link_flags:
        expected_inh_link_flags += b" " + b" ".join(new_inh_link_flags)

    # Don't touch the file if we don't need to, otherwise put the expected values into
    # 'contents' and write it into the la file.

    changed = False
    if dep_libs != expected_dep_libs:
        contents = contents.replace(
            b"dependency_libs='" + dep_libs + b"'",
            b"dependency_libs='" + expected_dep_libs + b"'",
        )
        changed = True

    if inh_link_flags is not None and expected_inh_link_flags != inh_link_flags:
        contents = contents.replace(
            b"inherited_linker_flags='" + inh_link_flags + b"'",
            b"inherited_linker_flags='" + expected_inh_link_flags + b"'",
        )
        changed = True

    if changed:
        return True, contents
    return False, None



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import tempfile

from portage import os
from portage import shutil
from portage import _unicode_encode
from portage.tests import TestCase
from portage.util import getconfig
from portage.exception import ParseError


class GetConfigTestCase(TestCase):
    """
    Test that getconfig() produces that same result as bash would when
    sourcing the same input.
    """

    _cases = {
        "FETCHCOMMAND": 'wget -t 3 -T 60 --passive-ftp -O "${DISTDIR}/${FILE}" "${URI}"',
        "FETCHCOMMAND_RSYNC": 'rsync -LtvP "${URI}" "${DISTDIR}/${FILE}"',
        "FETCHCOMMAND_SFTP": 'bash -c "x=\\${2#sftp://} ; host=\\${x%%/*} ; port=\\${host##*:} ; host=\\${host%:*} ; [[ \\${host} = \\${port} ]] && port= ; eval \\"declare -a ssh_opts=(\\${3})\\" ; exec sftp \\${port:+-P \\${port}} \\"\\${ssh_opts[@]}\\" \\"\\${host}:/\\${x#*/}\\" \\"\\$1\\"" sftp "${DISTDIR}/${FILE}" "${URI}" "${PORTAGE_SSH_OPTS}"',
        "FETCHCOMMAND_SSH": 'bash -c "x=\\${2#ssh://} ; host=\\${x%%/*} ; port=\\${host##*:} ; host=\\${host%:*} ; [[ \\${host} = \\${port} ]] && port= ; exec rsync --rsh=\\"ssh \\${port:+-p\\${port}} \\${3}\\" -avP \\"\\${host}:/\\${x#*/}\\" \\"\\$1\\"" rsync "${DISTDIR}/${FILE}" "${URI}" "${PORTAGE_SSH_OPTS}"',
        "PORTAGE_ELOG_MAILSUBJECT": "[portage] ebuild log for ${PACKAGE} on ${HOST}",
    }

    def testGetConfig(self):
        make_globals_file = os.path.join(self.cnf_path, "make.globals")
        d = getconfig(make_globals_file)
        for k, v in self._cases.items():
            self.assertEqual(d[k], v)

    def testGetConfigSourceLex(self):
        try:
            tempdir = tempfile.mkdtemp()
            make_conf_file = os.path.join(tempdir, "make.conf")
            with open(make_conf_file, "w") as f:
                f.write('source "${DIR}/sourced_file"\n')
            sourced_file = os.path.join(tempdir, "sourced_file")
            with open(sourced_file, "w") as f:
                f.write('PASSES_SOURCING_TEST="True"\n')

            d = getconfig(make_conf_file, allow_sourcing=True, expand={"DIR": tempdir})

            # PASSES_SOURCING_TEST should exist in getconfig result.
            self.assertTrue(d is not None)
            self.assertEqual("True", d["PASSES_SOURCING_TEST"])

            # With allow_sourcing=True and empty expand map, this should
            # throw a FileNotFound exception.
            self.assertRaisesMsg(
                "An empty expand map should throw an exception",
                ParseError,
                getconfig,
                make_conf_file,
                allow_sourcing=True,
                expand={},
            )
        finally:
            shutil.rmtree(tempdir)

    def testGetConfigProfileEnv(self):
        # Test the mode which is used to parse /etc/env.d and /etc/profile.env.

        cases = {
            "LESS_TERMCAP_mb": r"$\E[01;31m",  # bug #410625
        }

        with tempfile.NamedTemporaryFile(mode="wb") as f:
            # Format like env_update formats /etc/profile.env.
            for k, v in cases.items():
                if v.startswith("$") and not v.startswith("${"):
                    line = "export {}=$'{}'\n".format(k, v[1:])
                else:
                    line = "export {}='{}'\n".format(k, v)
                f.write(_unicode_encode(line))
            f.flush()

            d = getconfig(f.name, expand=False)
            for k, v in cases.items():
                self.assertEqual(d.get(k), v)
